(*BT
  Fredi Cobo
  Dic/00
  Btree manager*)

<*+M2EXTENSIONS*>
<*+M2ADDTYPES*>
<*ALIGNMENT="1"*>

IMPLEMENTATION MODULE bt;

FROM SYSTEM  IMPORT BYTE,ADR,ADDRESS;
FROM Storage IMPORT ALLOCATE,DEALLOCATE;
FROM Lib     IMPORT Fill,Move,AddAddr,IncAddr,DecAddr;
FROM Str     IMPORT Copy,Append,Prepend,Concat,Length,CardToStr,ItemS;
FROM FIO     IMPORT Size,Seek,WrBin,RdBin;
IMPORT FIO,Lib,Str,Windows,Win32IOChan;

CONST
IDXLEN = 8;
IDXSIZE= 65024;                  (*63.5 K              *)
IDXMAX = IDXSIZE DIV IDXLEN - 2; (*extra space to float*)

TYPE
QSEARCH= (foundBottom,foundTop,foundThis,notFoundFirst,notFoundLast,notFound);
PAGE   = ARRAY[0..5] OF BYTE;       (*0..1: pagHeap
                                      2..3: n
                                      4..5: k      *)

VAR
_c     : CARDINAL;
_lKey  : KEYITEM;
_hKey  : KEYITEM;
 idx   : ADDRESS;
 idxNum: CARDINAL;
 idxOfs: CARDINAL;
 idxDat: ARRAY[0..IDXSIZE-1] OF BYTE;

MODULE XDS;

FROM Lib         IMPORT GetTime;
FROM Windows     IMPORT HANDLE,LockFile,UnlockFile;
FROM Win32IOChan IMPORT GetSysHandle;
IMPORT FIO;

EXPORT Lock,Unlock;

CONST
POS    = 808;   (*pos.mod*)
TIMEOUT= 180;   (*seconds*)

  PROCEDURE Seconds(): LONGCARD;
  VAR
  Hrs,Mins,Secs,Hsecs: CARDINAL;
  BEGIN
    GetTime(Hrs,Mins,Secs,Hsecs);

    RETURN VAL(LONGCARD,Hrs )*3600
          +VAL(LONGCARD,Mins)*60
          +VAL(LONGCARD,Secs)
  END Seconds;

  PROCEDURE Delay;
  VAR
  start: LONGCARD;
  BEGIN
    start:=Seconds();
    LOOP
      IF Seconds()#start THEN
         RETURN
      END
    END
  END Delay;

  PROCEDURE Error(msg: ARRAY OF CHAR);
  CONST
  FLE= "BT.TXT";

  VAR
  h: FIO.File;
  BEGIN
    IF FIO.Exists(FLE) THEN
       h:=FIO.Append(FLE)
    ELSE
       h:=FIO.Create(FLE)
    END;

    FIO.WrStr(h,msg);FIO.WrLn(h);FIO.Close(h);HALT
  END Error;

  PROCEDURE Lock(h: FIO.File);
  VAR
  handle: HANDLE;
  cnt   : CARDINAL;
  BEGIN
    handle:=GetSysHandle(h);

    cnt:=0;
    LOOP
      IF LockFile(handle,POS,0,1,0) THEN
         RETURN
      END;

      Delay;

      INC(cnt);
      IF cnt=TIMEOUT THEN
         Error("bt/XDS/Lock/1")
      END
    END
  END Lock;

  PROCEDURE Unlock(h: FIO.File);
  BEGIN
    IF NOT UnlockFile(GetSysHandle(h),POS,0,1,0) THEN
       Error("bt/XDS/Unlock/1")
    END
  END Unlock;

BEGIN
END XDS;

PROCEDURE Error(msg: ARRAY OF CHAR);
CONST
FLE= "BT.TXT";

VAR
h: FIO.File;
BEGIN
  IF FIO.Exists(FLE) THEN
     h:=FIO.Append(FLE)
  ELSE
     h:=FIO.Create(FLE)
  END;

  FIO.WrStr(h,msg);FIO.WrLn(h);FIO.Close(h);HALT
END Error;

PROCEDURE _Trim(VAR txt: ARRAY OF CHAR);
VAR
k,w,l,r,j: CARDINAL;
ch       : CHAR;
filled   : BOOLEAN;
BEGIN
     IF Str.Compare(txt,"" )=0 THEN Copy(txt,"");RETURN
  ELSIF Str.Compare(txt," ")=0 THEN Copy(txt,"");RETURN
  ELSIF HIGH(txt)=0            THEN              RETURN
    END;

  txt[HIGH(txt)]:=CHR(EOS);ch:=CHR(_c);_c:=BLANK;

  w:=Length(txt);
  IF w=0 THEN
     txt[0]:=CHR(EOS);RETURN
  END;

  filled:=TRUE;
  FOR k:=0 TO w-1 DO
    filled:=filled AND (txt[k]=ch)
  END;
  IF filled THEN
     txt[0]:=CHR(EOS);RETURN
  END;

  (*left pointer*)
  l:=0;
  LOOP
    IF txt[l]#ch THEN EXIT END;
    INC(l)
  END;

  (*right pointer*)
  r:=w-1;
  LOOP
    IF txt[r]#ch THEN EXIT END;
    DEC(r)
  END;

  (*copy string*)
  j:=0;
  FOR k:=l TO r DO
    txt[j]:=txt[k];INC(j)
  END;
  txt[j]:=CHR(EOS)
END _Trim;

PROCEDURE _Pad(VAR txt: ARRAY OF CHAR;
                   w  : CARDINAL);
VAR
k,h: CARDINAL;
ch : CHAR;
BEGIN
  h:=HIGH(txt);txt[h]:=CHR(EOS);ch:=CHR(_c);_c:=BLANK;

     IF w>h THEN w:=h
  ELSIF w=0 THEN txt[0]:=CHR(EOS);RETURN
    END;

  FOR k:=Length(txt) TO w-1 DO
      txt[k]:=ch
  END;
  txt[w]:=CHR(EOS)
END _Pad;

PROCEDURE _Remove(VAR txt: ARRAY OF CHAR);
VAR
j,k: CARDINAL;
ch : CHAR;
BEGIN
  txt[HIGH(txt)]:=CHR(EOS);ch:=CHR(_c);_c:=BLANK;

  IF Length(txt)=0 THEN RETURN END;

  j:=0;
  FOR k:=0 TO Length(txt)-1 DO
      IF txt[k]#ch THEN txt[j]:=txt[k];INC(j) END
  END;
  txt[j]:=CHR(EOS)
END _Remove;

PROCEDURE Null(s: ARRAY OF CHAR): BOOLEAN;
BEGIN
  _Trim(s);RETURN s[0]=CHR(EOS)
END Null;

PROCEDURE fConcat(s1,s2: ARRAY OF CHAR): STR;
VAR
txt: STR;
BEGIN
  Concat(txt,s1,s2);RETURN txt
END fConcat;

PROCEDURE Pack(n: LONGCARD): STR;
VAR
ok : BOOLEAN;
txt: STR;
BEGIN
  CardToStr(n,txt,10,ok);_Trim(txt);RETURN txt
END Pack;

PROCEDURE DPack(n1,n2: LONGCARD): STR;
VAR
ok           : BOOLEAN;
txt1,txt2,txt: STR;
BEGIN
  CardToStr(n1,txt1,10,ok);_Trim(txt1);
  CardToStr(n2,txt2,10,ok);_Trim(txt2);

  Concat(txt,txt1,"/");Append(txt,txt2);RETURN txt
END DPack;

PROCEDURE ErrFirma(headFirma: FIRMA): BOOLEAN;
VAR
k: CARDINAL;
BEGIN
  FOR k:=0 TO HIGH(firma) DO
    IF headFirma[k]#firma[k] THEN
       RETURN TRUE
    END
  END;

  RETURN FALSE
END ErrFirma;

PROCEDURE Make(keyAdd,recAdd: ADDRESS;
               make         : MAKE);
VAR
k,pos,data: CARDINAL;

  PROCEDURE Byte;
  BEGIN
    Move(AddAddr(recAdd,pos),keyAdd,1);IncAddr(keyAdd,1)
  END Byte;

  PROCEDURE Chars;
  BEGIN
    Move(AddAddr(recAdd,pos),keyAdd,data);IncAddr(keyAdd,data)
  END Chars;

  PROCEDURE String;
  VAR
  s: ARRAY[0..KEYSIZE] OF CHAR;
  BEGIN
    Move(AddAddr(recAdd,pos),ADR(s),data);s[data]:=CHR(EOS);
    _Remove(s);_Pad(s,data);
    Move(ADR(s),keyAdd,data);IncAddr(keyAdd,data)
  END String;

  PROCEDURE Digits;
  VAR
  i : ADDRESS;
  ch: CHAR;
  j : CARDINAL;
  n : LONGCARD;
  BEGIN
    i:=AddAddr(recAdd,pos);n:=0;
    FOR j:=1 TO data DO
      Move(i,ADR(ch),1);IncAddr(i,1);

      IF (ORD(ch)>=48) AND (ORD(ch)<=57) THEN
         Move(ADR(ch),keyAdd,1);IncAddr(keyAdd,1);INC(n)
      END
    END;

    IF n#make[k].size THEN
       Error("bt/Make/Digits/1 - wrong size")
    END
  END Digits;

  PROCEDURE Longcard;
  VAR
  i: ADDRESS;
  BEGIN
    i:=AddAddr(recAdd,pos+3);

    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1)
  END Longcard;

  PROCEDURE Reverse;
  VAR
  i,datAdd: ADDRESS;
  dat     : LONGCARD;
  BEGIN
    datAdd:=ADR(dat);

    Move(AddAddr(recAdd,pos),datAdd,4);dat:=MAX(LONGCARD)-dat;

    i:=AddAddr(datAdd,3);

    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1)
  END Reverse;

  PROCEDURE Seq;
  VAR
  i: ADDRESS;
  BEGIN
    i:=AddAddr(recAdd,2);

    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1);DecAddr(i,1);
    Move(i,keyAdd,1);IncAddr(keyAdd,1)
  END Seq;

BEGIN
  k:=0;
  LOOP
       IF      k      =KEYFIELDS THEN RETURN
    ELSIF make[k].type=KEYEND    THEN RETURN
      END;

    pos :=VAL(CARDINAL,make[k].pos );
    data:=VAL(CARDINAL,make[k].data);

    CASE make[k].type OF
       "B": Byte
      |"X": Chars
      |"S": String
      |"D": Digits
      |"L": Longcard
      |"R": Reverse
      |"Q": Seq
    ELSE
            Error(fConcat("bt/Make/1 - wrong type of key: ",make[k].type))
    END;

    INC(k)
  END
END Make;

(*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*)

PROCEDURE Reset;
BEGIN
  id:=0;first:=NIL;last:=NIL
END Reset;

PROCEDURE File(id: LONGCARD): NodePtr;
VAR
p: NodePtr;
BEGIN
  p:=first;
  WHILE p#NIL DO
    IF p^.id=id THEN
       RETURN p
    END;

    p:=p^.next
  END;

  Error(fConcat("bt/File/1 - file id not found: ",Pack(id)))
END File;

PROCEDURE Read(    h    : FIO.File;
               VAR data : ARRAY OF BYTE;
                   count: CARDINAL);
BEGIN
  IF RdBin(h,data,count)#count THEN
     Error("bt/Read/1")
  END
END Read;

PROCEDURE PadKey(VAR key  : KEYITEM;
                     pos  : CARDINAL;
                     value: BYTE);
VAR
k: CARDINAL;
BEGIN
  FOR k:=pos TO KEYSIZE-1 DO
    key[k]:=value
  END
END PadKey;

PROCEDURE Compare(VAR source,dest: ARRAY OF BYTE;
                      len        : CARDINAL): CompareType;
VAR
pos: CARDINAL;
BEGIN
  pos:=Lib.Compare(ADR(source),ADR(dest),len);

     IF pos=len                         THEN RETURN equal
  ELSIF ORD(source[pos])>ORD(dest[pos]) THEN RETURN greater
   ELSE                                      RETURN lower
    END
END Compare;

PROCEDURE ZipPage(VAR page   : PAGE;
                      pagHeap: LONGCARD;
                      n      : LONGCARD;
                      k      : LONGCARD);

  PROCEDURE Card(n: LONGCARD;
                 i: CARDINAL);
  VAR
  dat: ARRAY[0..3] OF BYTE;
  BEGIN
    Move(ADR(n),ADR(dat),4);

    page[i  ]:=dat[1];
    page[i+1]:=dat[0]
  END Card;

BEGIN
  Card(pagHeap,0);
  Card(n      ,2);
  Card(k      ,4)
END ZipPage;

PROCEDURE UnzipPage(    page   : PAGE;
                    VAR pagHeap: CARDINAL;
                    VAR n      : CARDINAL;
                    VAR k      : CARDINAL);

  PROCEDURE Card(i: CARDINAL): CARDINAL;
  BEGIN
    RETURN ORD(page[i])*256 + ORD(page[i+1])
  END Card;

BEGIN
  pagHeap:=Card(0);
  n      :=Card(2);
  k      :=Card(4)
END UnzipPage;

PROCEDURE NdxLen(size: CARDINAL): CARDINAL;
BEGIN
  RETURN  size   (*lkey*)
         +size   (*hkey*)
         +4      (*pos *)
         +4      (*ofs *)
         +4      (*n   *)
END NdxLen;

PROCEDURE NdxMax(size: CARDINAL): CARDINAL;
BEGIN
  RETURN NDXHEAP DIV NdxLen(size)
END NdxMax;

PROCEDURE PagLen(size: CARDINAL): CARDINAL;
BEGIN
  RETURN  size   (*key*)
         +4      (*pos*)
END PagLen;

PROCEDURE GetNdx(    i   : ADDRESS;
                     size: CARDINAL;
                 VAR lkey: ARRAY OF BYTE;
                 VAR hkey: ARRAY OF BYTE;
                 VAR pos : LONGCARD;
                 VAR ofs : LONGCARD;
                 VAR n   : LONGCARD);
BEGIN
  Move(i,ADR(lkey),size);IncAddr(i,size);
  Move(i,ADR(hkey),size);IncAddr(i,size);
  Move(i,ADR(pos ),4   );IncAddr(i,4   );
  Move(i,ADR(ofs ),4   );IncAddr(i,4   );
  Move(i,ADR(n   ),4   )
END GetNdx;

PROCEDURE xGetNdx(    i      : ADDRESS;
                      size   : CARDINAL;
                  VAR loaded : BOOLEAN;
                  VAR pag    : ADDRESS;
                  VAR pagHeap: CARDINAL;
                  VAR n      : CARDINAL;
                  VAR k      : CARDINAL);
VAR
page: PAGE;
BEGIN
                                 IncAddr(i,size); (*lkey*)
                                 IncAddr(i,size); (*hkey*)
  Move(i,ADR(loaded),     1    );IncAddr(i,1   );
  Move(i,ADR(pag   ),     4    );IncAddr(i,4   );
  Move(i,ADR(page  ),SIZE(PAGE));

  UnzipPage(page,pagHeap,n,k)
END xGetNdx;

PROCEDURE SetNdx(i   : ADDRESS;
                 size: CARDINAL;
                 lkey: ARRAY OF BYTE;
                 hkey: ARRAY OF BYTE;
                 pos : LONGCARD;
                 ofs : LONGCARD;
                 n   : LONGCARD);
BEGIN
  Move(ADR(lkey),i,size);IncAddr(i,size);
  Move(ADR(hkey),i,size);IncAddr(i,size);
  Move(ADR(pos ),i,4   );IncAddr(i,4   );
  Move(ADR(ofs ),i,4   );IncAddr(i,4   );
  Move(ADR(n   ),i,4   )
END SetNdx;

PROCEDURE xSetNdx(i      : ADDRESS;
                  size   : CARDINAL;
                  loaded : BOOLEAN;
                  pag    : ADDRESS;
                  pagHeap: LONGCARD;
                  n      : LONGCARD;
                  k      : LONGCARD);
VAR
page: PAGE;
BEGIN
  ZipPage(page,pagHeap,n,k);

  Move(ADR(loaded),AddAddr(i,2*size  ),     1    );
  Move(ADR(pag   ),AddAddr(i,2*size+1),     4    );
  Move(ADR(page  ),AddAddr(i,2*size+5),SIZE(PAGE))
END xSetNdx;

PROCEDURE GetPag(    i   : ADDRESS;
                     size: CARDINAL;
                 VAR key : ARRAY OF BYTE;
                 VAR pos : LONGCARD);
BEGIN
  Move(i,ADR(key),size);IncAddr(i,size);
  Move(i,ADR(pos),4   )
END GetPag;

PROCEDURE xGetPag(    i   : ADDRESS;
                      size: CARDINAL;
                  VAR dat : ADDRESS);
BEGIN
                    ;IncAddr(i,size);  (*key*)
  Move(i,ADR(dat),4)
END xGetPag;

PROCEDURE SetPag(i   : ADDRESS;
                 size: CARDINAL;
                 key : ARRAY OF BYTE;
                 pos : LONGCARD);
BEGIN
  Move(ADR(key),i,size);IncAddr(i,size);
  Move(ADR(pos),i,4   )
END SetPag;

PROCEDURE xSetPag(i   : ADDRESS;
                  size: CARDINAL;
                  pos : LONGCARD);
BEGIN
                        IncAddr(i,size);
  Move(ADR(pos),i,4   )
END xSetPag;

PROCEDURE GetIdx(    i      : ADDRESS;
                 VAR address: ADDRESS;
                 VAR pos    : LONGCARD);
BEGIN
  Move(i,ADR(address),4);IncAddr(i,4);
  Move(i,ADR(pos    ),4)
END GetIdx;

PROCEDURE SetIdx(i      : ADDRESS;
                 address: ADDRESS;
                 pos    : LONGCARD);
BEGIN
  Move(ADR(address),i,4);IncAddr(i,4);
  Move(ADR(pos    ),i,4)
END SetIdx;

PROCEDURE RdNdx(    h     : FIO.File;
                    ndxPos: LONGCARD;
                    ndxOfs: LONGCARD;
                    loc   : CARDINAL;
                    size  : CARDINAL;
                VAR lkey  : ARRAY OF BYTE;
                VAR hkey  : ARRAY OF BYTE;
                VAR pos   : LONGCARD;
                VAR ofs   : LONGCARD;
                VAR n     : LONGCARD);
VAR
ndx: ADDRESS;
len: CARDINAL;
BEGIN
  len:=NdxLen(size);ALLOCATE(ndx,len);

  Seek(h,ndxPos+ndxOfs + VAL(LONGCARD,loc)*VAL(LONGCARD,len));
  Read(h,ndx^,len);

  GetNdx(ndx,size,lkey,hkey,pos,ofs,n);

  DEALLOCATE(ndx,len)
END RdNdx;

PROCEDURE WrNdx(h     : FIO.File;
                ndxPos: LONGCARD;
                ndxOfs: LONGCARD;
                loc   : CARDINAL;
                size  : CARDINAL;
                lkey  : ARRAY OF BYTE;
                hkey  : ARRAY OF BYTE;
                pos   : LONGCARD;
                ofs   : LONGCARD;
                n     : LONGCARD);
VAR
ndx: ADDRESS;
len: CARDINAL;
BEGIN
  len:=NdxLen(size);ALLOCATE(ndx,len);

  SetNdx(ndx,size,lkey,hkey,pos,ofs,n);

  Seek(h,ndxPos+ndxOfs + VAL(LONGCARD,loc)*VAL(LONGCARD,len));
  WrBin(h,ndx^,len);

  DEALLOCATE(ndx,len)
END WrNdx;

PROCEDURE ResetIdx;
BEGIN
  idxNum:=0
END ResetIdx;

PROCEDURE AddIdx(address: ADDRESS;
                 pos    : LONGCARD);
TYPE
SEARCH= (empty,full,before,after,inside);

VAR
idxLoc: CARDINAL;

  PROCEDURE SearchIdx(n: CARDINAL): SEARCH;
  VAR
  lptr,hptr,ptr,m: CARDINAL;
  cmp            : CompareType;

    PROCEDURE Key(loc: CARDINAL);
    VAR
    _pos: LONGCARD;
    _   : ADDRESS;
    BEGIN
      GetIdx(AddAddr(idx,idxOfs + loc*IDXLEN),_,_pos);

         IF pos=_pos THEN cmp:=equal
      ELSIF pos>_pos THEN cmp:=greater
       ELSE               cmp:=lower
        END;

      idxLoc:=loc
    END Key;

  BEGIN
    (*casos especiales, tamao de la lista*)
       IF n =0      THEN RETURN empty
    ELSIF n>=IDXMAX THEN RETURN full
      END;

    (*caso especial item 0*)
    Key(0);

       IF cmp=equal THEN Error("bt/AddIdx/1")
    ELSIF cmp=lower THEN RETURN before
      END;

    (*caso especial item n-1*)
    Key(n-1);

       IF cmp=equal   THEN Error("bt/AddIdx/2")
    ELSIF cmp=greater THEN RETURN after
      END;

    (*situacion general*)
    lptr:=0;
    hptr:=n-1;
    LOOP
      ptr:=lptr + (hptr-lptr) DIV 2;
      m  :=hptr-lptr+1;

      (*item central*)
      Key(ptr);

         IF cmp=equal THEN Error("bt/AddIdx/3")
      ELSIF cmp=lower THEN hptr:=ptr
      ELSIF m  =2     THEN RETURN inside
       ELSE                lptr:=ptr
        END
    END
  END SearchIdx;

  PROCEDURE Underflow(): BOOLEAN;
  BEGIN
    RETURN idxOfs<IDXLEN
  END Underflow;

  PROCEDURE Overflow(): BOOLEAN;
  BEGIN
    RETURN idxOfs + (idxNum+1)*IDXLEN > IDXSIZE
  END Overflow;

  PROCEDURE Relocate;
  VAR
  _idxOfs: CARDINAL;
  BEGIN
   _idxOfs:=idxOfs;
    idxOfs:=(IDXSIZE - idxNum*IDXLEN) DIV 2;

    Move(AddAddr(idx,_idxOfs),AddAddr(idx,idxOfs),idxNum*IDXLEN)
  END Relocate;

  PROCEDURE Init;
  BEGIN
    idxOfs:=IDXSIZE DIV 2;

    SetIdx(AddAddr(idx,idxOfs),address,pos)
  END Init;

  PROCEDURE Before;
  BEGIN
    IF Underflow() THEN
       Relocate
    END;

    DEC(idxOfs,IDXLEN);
    SetIdx(AddAddr(idx,idxOfs),address,pos)
  END Before;

  PROCEDURE After;
  BEGIN
    IF Overflow() THEN
       Relocate
    END;

    SetIdx(AddAddr(idx,idxOfs + idxNum*IDXLEN),address,pos)
  END After;

  PROCEDURE Inside;

    PROCEDURE High;
    VAR
    source,dest: ADDRESS;
    BEGIN
      source:=AddAddr(idx,idxOfs + (idxLoc+1)*IDXLEN);
      dest  :=AddAddr(idx,idxOfs + (idxLoc+2)*IDXLEN);

        Move(source,dest,(idxNum-idxLoc-1)*IDXLEN);
      SetIdx(source,address,pos)
    END High;

    PROCEDURE Low;
    BEGIN
        Move(AddAddr(idx,idxOfs),AddAddr(idx,idxOfs-IDXLEN),(idxLoc+1)*IDXLEN);
      SetIdx(AddAddr(idx,idxOfs+idxLoc*IDXLEN),address,pos);

      DEC(idxOfs,IDXLEN)
    END Low;

  BEGIN
    IF Underflow() OR Overflow() THEN
       Relocate
    END;

    IF idxLoc+1 > idxNum-idxLoc-1 THEN
       High
    ELSE
       Low
    END
  END Inside;

BEGIN
  CASE SearchIdx(idxNum) OF
     full  : Error("bt/AddIdx/1")
    |empty : Init
    |before: Before
    |after : After
    |inside: Inside
  ELSE
             Error("bt/AddIdx/2")
  END;

  INC(idxNum)
END AddIdx;

PROCEDURE QCommon(    id    : LONGCARD;
                  VAR h     : FIO.File;
                  VAR q     : QUERY;
                      index : IndexType;
                      load  : BOOLEAN;
                      lock  : BOOLEAN): BOOLEAN;
VAR
file: NodePtr;
head: HEAD;
key : KEY;
BEGIN
  (*get item from list*)
    file   :=             File(id)      ;
    h      :=             file^.h       ;
  q.h      :=             file^.h       ;
  q.recSize:=VAL(CARDINAL,file^.recSize);

  (*read head*)
  IF lock THEN
     Lock(h)
  END;

  Seek(h,0);Read(h,head,SIZE(head));

  CASE index OF
      seq: key:=head.seq;q.ndxPos:=SEQPOS;q.make:=      seqMake
    |main: key:=head.ndx;q.ndxPos:=NDXPOS;q.make:=file^.ndxMake
    | aux: key:=head.aux;q.ndxPos:=AUXPOS;q.make:=file^.auxMake;
           IF head.aux.size=0 THEN
              Error("bt/QCommon/1 - bad index (aux)")
           END
  ELSE
           Error("bt/QCommon/2 - bad index")
  END;

  q.PAGHEAP:=head.pagHeap;
  q.file   :=file;
  q.ndx    :=NIL;
  q.ndxOfs :=key.ofs;
  q.ndxSize:=VAL(CARDINAL,key.size);
  q.ndxLen :=NdxLen(VAL(CARDINAL,key.size));
  q.ndxHeap:=VAL(CARDINAL,key.n)*q.ndxLen;
  q.pagLen :=PagLen(VAL(CARDINAL,key.size));
  q.pagMax :=VAL(CARDINAL,head.pagHeap) DIV q.pagLen;
  q.n      :=VAL(CARDINAL,key.n);
  q.K      :=0;
  q.k      :=0;
  q.recs   :=0;
  q.aux    :=head.aux.size#0;
  q.guard  :=GUARD;

     IF (key.n=0) AND lock THEN Unlock(h);RETURN FALSE
  ELSIF  key.n=0           THEN           RETURN FALSE
  ELSIF  NOT load          THEN           RETURN TRUE
    END;

  ALLOCATE(q.ndx,q.ndxHeap);

  IF file^.query#NIL THEN
     Error("bt/QCommon/3 - query already exists")
  END;

  file^.query:=ADR(q);

  Seek(h,q.ndxPos+q.ndxOfs);Read(h,q.ndx^,q.ndxHeap);

  RETURN TRUE
END QCommon;

PROCEDURE QSort(VAR q    : QUERY;
                    i    : ADDRESS;
                    first: CARDINAL;
                    last : CARDINAL);
VAR
k        : CARDINAL;
pos,ofs,n: LONGCARD;
lkey,hkey: KEYITEM;
BEGIN
  ResetIdx;
  FOR k:=first TO last DO
    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);AddIdx(i,pos+ofs);
    IncAddr(i,q.ndxLen);INC(q.recs,n)
  END
END QSort;

PROCEDURE QRead(q: QUERY);
VAR
pag,i,link: ADDRESS;
pagHeap,k : CARDINAL;
pos,_,__,n: LONGCARD;
___,____  : KEYITEM;
BEGIN
  i:=AddAddr(idx,idxOfs);
  FOR k:=1 TO idxNum DO
    GetIdx(i,link,pos);GetNdx(link,q.ndxSize,___,____,_,__,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;

    ALLOCATE(pag,pagHeap);
    IF pag=NIL THEN
       Error("bt/QRead/1 - not enough memory")
    END;

    Seek(q.h,pos);Read(q.h,pag^,pagHeap);

    xSetNdx(link,q.ndxSize,FALSE,pag,VAL(LONGCARD,pagHeap),n,0);

    IncAddr(i,IDXLEN)
  END
END QRead;

PROCEDURE QSearch(    q        : QUERY;
                      key      : KEYITEM;
                  VAR pos,ofs,n: LONGCARD;
                  VAR ndxLoc   : CARDINAL;
                      disk     : BOOLEAN): QSEARCH;
VAR
lptr,hptr,ptr,m: CARDINAL;
lcmp,hcmp      : CompareType;

  PROCEDURE Keys(loc: CARDINAL);
  BEGIN
    IF disk THEN
       RdNdx(q.h,q.ndxPos,q.ndxOfs,loc   ,q.ndxSize,_lKey,_hKey,pos,ofs,n)
    ELSE
       GetNdx(AddAddr(q.ndx,loc*q.ndxLen),q.ndxSize,_lKey,_hKey,pos,ofs,n)
    END;

    lcmp:=Compare(key,_lKey,q.ndxSize);
    hcmp:=Compare(key,_hKey,q.ndxSize);

    ndxLoc:=loc
  END Keys;

BEGIN
  (*caso especial item 0*)
  Keys(0);

     IF  lcmp=lower                       THEN RETURN notFoundFirst
  ELSIF (lcmp=greater) AND (hcmp=lower  ) THEN RETURN foundThis
  ELSIF                     hcmp=greater  THEN
  ELSIF (lcmp=greater) AND (hcmp=equal  ) THEN RETURN foundTop
  ELSIF (lcmp=equal  ) AND (hcmp=lower  ) THEN RETURN foundBottom
  ELSIF (lcmp=equal  ) AND (hcmp=equal  ) THEN RETURN foundBottom
    END;

  (*caso especial item n-1*)
  Keys(q.n-1);

     IF  lcmp=lower                       THEN
  ELSIF (lcmp=greater) AND (hcmp=lower  ) THEN RETURN foundThis
  ELSIF                     hcmp=greater  THEN RETURN notFoundLast
  ELSIF (lcmp=greater) AND (hcmp=equal  ) THEN RETURN foundTop
  ELSIF (lcmp=equal  ) AND (hcmp=lower  ) THEN RETURN foundBottom
  ELSIF (lcmp=equal  ) AND (hcmp=equal  ) THEN RETURN foundBottom
    END;

  (*situacion general*)
  lptr:=0;
  hptr:=q.n-1;
  LOOP
    ptr:=lptr + (hptr-lptr) DIV 2;
    m  :=hptr-lptr+1;

    (*item central*)
    Keys(ptr);

       IF  lcmp=lower                                 THEN hptr:=ptr
    ELSIF (lcmp=greater) AND (hcmp=lower  )           THEN RETURN foundThis
    ELSIF                    (hcmp=greater) AND (m=2) THEN RETURN notFound
    ELSIF                     hcmp=greater            THEN lptr:=ptr
    ELSIF (lcmp=greater) AND (hcmp=equal  )           THEN RETURN foundTop
    ELSIF (lcmp=equal  ) AND (hcmp=lower  )           THEN RETURN foundBottom
    ELSIF (lcmp=equal  ) AND (hcmp=equal  )           THEN RETURN foundBottom
     ELSE                                                  Error("bt/QSearch/1")
      END
  END
END QSearch;

PROCEDURE QPages(VAR q         : QUERY;
                     pag,i     : ADDRESS;
                     pagHeap   : CARDINAL;
                     n         : LONGCARD;
                     first,last: CARDINAL);
VAR
_pag       : ADDRESS;
_pagHeap,_n: CARDINAL;
BEGIN
  _n:=last-first+1;
  IF _n=VAL(CARDINAL,n) THEN
     xSetNdx(i,q.ndxSize,FALSE,pag,VAL(LONGCARD,pagHeap),n,0)
  ELSE
     _pagHeap:=_n*q.pagLen;
     ALLOCATE(_pag,_pagHeap);

     Move(AddAddr(pag,first*q.pagLen),_pag,_pagHeap);

     xSetNdx(i,q.ndxSize,FALSE,_pag,VAL(LONGCARD,_pagHeap),VAL(LONGCARD,_n),0);

     DEALLOCATE(pag,pagHeap)
  END;

  INC(q.recs,VAL(LONGCARD,_n))
END QPages;

PROCEDURE QSearchPag(    q  : QUERY;
                         pag: ADDRESS;
                         key: KEYITEM;
                         len: CARDINAL;
                         n  : CARDINAL;
                     VAR pos: LONGCARD;
                     VAR loc: CARDINAL): BOOLEAN;
VAR
lptr,hptr,ptr,m: CARDINAL;
cmp            : CompareType;

  PROCEDURE Key(k: CARDINAL);
  VAR
  _key: KEYITEM;
  BEGIN
    loc:=k;GetPag(AddAddr(pag,loc*q.pagLen),q.ndxSize,_key,pos);

    cmp:=Compare(key,_key,len)
  END Key;

BEGIN
  (*caso especial item 0*)
  Key(0);

     IF cmp=equal THEN RETURN TRUE
  ELSIF cmp=lower THEN RETURN FALSE
    END;

  (*caso especial item n-1*)
  Key(n-1);

     IF cmp=equal   THEN RETURN TRUE
  ELSIF cmp=greater THEN RETURN FALSE
    END;

  (*situacion general*)
  lptr:=0;
  hptr:=n-1;
  LOOP
    ptr:=lptr + (hptr-lptr) DIV 2;
    m  :=hptr-lptr+1;

    (*item central*)
    Key(ptr);

       IF cmp=equal THEN RETURN TRUE
    ELSIF cmp=lower THEN hptr:=ptr
    ELSIF m  =2     THEN RETURN FALSE
     ELSE                lptr:=ptr
      END
  END
END QSearchPag;

PROCEDURE Free(id: LONGCARD);
VAR
p: NodePtr;

  PROCEDURE FreeQuery(q: QueryPtr);
  BEGIN
       IF q=NIL          THEN
    ELSIF q^.guard#GUARD THEN Error("bt/Free/FreeQuery/1 - query corrupted")
     ELSE                     CloseQuery(q^)
      END
  END FreeQuery;

BEGIN
  p:=first;
  WHILE p#NIL DO
       IF p^.id#id                THEN
    ELSIF (p=first) AND (p=last)  THEN
          FreeQuery(p^.query);DEALLOCATE(p,SIZE(p^));Reset;RETURN
    ELSIF  p=first                THEN
          p^.next^.prev:=NIL;
          first        :=p^.next;
          FreeQuery(p^.query);DEALLOCATE(p,SIZE(p^));RETURN
    ELSIF                p=last   THEN
          p^.prev^.next:=NIL;
          last         :=p^.prev;
          FreeQuery(p^.query);DEALLOCATE(p,SIZE(p^));RETURN
     ELSE
          p^.prev^.next:=p^.next;
          p^.next^.prev:=p^.prev;
          FreeQuery(p^.query);DEALLOCATE(p,SIZE(p^));RETURN
      END;

    p:=p^.next
  END;

  Error(fConcat("bt/Free/1 - file id not found: ",Pack(id)))
END Free;

PROCEDURE Create(name : ARRAY OF CHAR;
                 head : HEAD;
                 extra: CARDINAL);
VAR
h   : FIO.File;
ndx : ADDRESS;
lock: LOCK;

  PROCEDURE Key(VAR key      : KEY;
                    indexType: IndexType);

    PROCEDURE Size(make: MAKE): LONGCARD;
    VAR
    size: LONGCARD;
    k   : CARDINAL;
    BEGIN
      IF indexType=seq THEN
         key.make:=seqMake;RETURN SEQSIZE
      END;

      k:=0;size:=0;
      LOOP
           IF      k      =KEYFIELDS THEN RETURN size
        ELSIF make[k].type=KEYEND    THEN RETURN size
          END;

           IF make[k].size=0            THEN Error("bt/Create/Key/Size/1 - null size"             )
        ELSIF make[k].size>KEYSIZE      THEN Error("bt/Create/Key/Size/2 - size overflow"         )
        ELSIF make[k].size>make[k].data THEN Error("bt/Create/Key/Size/3 - size greater than data")
          END;

        INC(size,make[k].size);INC(k)
      END
    END Size;

  BEGIN
    key.size:=Size(key.make);
    key.ofs :=0;
    key.n   :=0;

       IF (indexType=main) AND (key.size=0      ) THEN Error("bt/Create/Key/1 - null ndxSize"    )
    ELSIF (indexType=main) AND (key.size>KEYSIZE) THEN Error("bt/Create/Key/2 - ndxSize overflow")
    ELSIF (indexType=aux ) AND (key.size>KEYSIZE) THEN Error("bt/Create/Key/3 - auxSize overflow")
      END
  END Key;

BEGIN
     IF  Null(name)                             THEN Error("bt/Create/1 - null name"       )
  ELSIF  head.pagHeap=0                         THEN Error("bt/Create/2 - null pagHeap"    )
  ELSIF  head.pagHeap>SEQHEAP                   THEN Error("bt/Create/3 - pagHeap overflow")
  ELSIF (head.recSize=0      ) AND head.fixSize THEN Error("bt/Create/4 - null recSize"    )
    END;

  (*extra head info*)
  head.firma  :=firma;
  head.version:=VERSION;
  head.empty  :=TRUE;
  head.full   :=FALSE;
  head.counter:=1;
  head.n      :=0;

  Key(head.seq,seq );
  Key(head.ndx,main);
  Key(head.aux,aux );

  (*disk create*)
  ALLOCATE(ndx,NDXHEAP);Fill(ndx,NDXHEAP,0);

  Fill(ADR(lock),SIZE(LOCK),0);lock[SIZE(LOCK)-1]:=BYTE(ORD(""));

  h:=FIO.Create(name);

  WrBin(h,head,SIZE(HEAD));
  WrBin(h,lock,SIZE(LOCK));

  WrBin(h,ndx^,NDXHEAP);     (*SEQ  *)
  WrBin(h,ndx^,NDXHEAP);     (*NDX  *)
  IF head.aux.size#0 THEN
  WrBin(h,ndx^,NDXHEAP)      (*AUX  *)
  END;

  IF extra#0 THEN            (*EXTRA*)
     WrBin(h,ndx^,extra)
  END;

  FIO.Close(h);

  DEALLOCATE(ndx,NDXHEAP)
END Create;

PROCEDURE Open(name   : ARRAY OF CHAR;
               recSize: LONGCARD): LONGCARD;
VAR
h   : FIO.File;
sign: SIGN;
head: HEAD;
curr: NodePtr;
BEGIN
  (*error trap*)
     IF Null(name)           THEN Error(        "bt/Open/1 - null name"             )
  ELSIF NOT FIO.Exists(name) THEN Error(fConcat("bt/Open/2 - file not found: ",name))
    END;

  (*open file*)
  h:=FIO.Open(name);

  Seek(h,0);Read(h,head,SIZE(HEAD));
  Move(ADR(head),ADR(sign),SIZE(SIGN));

     IF  ErrFirma(sign.firma)                   THEN FIO.Close(h);Error(        "bt/Open/3 - bad signature"                                   )
  ELSIF  sign.version#VERSION                   THEN FIO.Close(h);Error(fConcat("bt/Open/4 - bad version: "      , Pack(sign.version        )))
  ELSIF (head.recSize#recSize) AND head.fixSize THEN FIO.Close(h);Error(fConcat("bt/Open/5 - wrong record size: ",DPack(head.recSize,recSize)))
    END;

  (*add item to list*)
  NEW(curr);INC(id);

  curr^.id     :=id;
  curr^.h      :=h;
  curr^.recSize:=recSize;
  curr^.ndxMake:=head.ndx.make;
  curr^.auxMake:=head.aux.make;
  curr^.query  :=NIL;

  IF first=NIL THEN
     curr^.prev:=NIL;
     curr^.next:=NIL;
     first     :=curr;
     last      :=curr
  ELSE
     last^.next:=curr;
     curr^.prev:=last;
     curr^.next:=NIL;
     last      :=curr
  END;

  RETURN id
END Open;

PROCEDURE Close(id: LONGCARD);
BEGIN
  FIO.Close(File(id)^.h);Free(id)
END Close;

PROCEDURE UnLock(h: FIO.File);
BEGIN
  Unlock(h)
END UnLock;

PROCEDURE Add(    id : LONGCARD;
              VAR dat: ARRAY OF BYTE): BOOLEAN;
VAR
h              : FIO.File;
file           : NodePtr;
head           : HEAD;
ndxMake,auxMake: MAKE;
recSize,recPos : LONGCARD;

  PROCEDURE WrDat;
  BEGIN
    Move(ADR(head.n),ADR(dat),4); (*seq  number*)

    recPos:=Size(h);Seek(h,recPos);WrBin(h,dat,VAL(CARDINAL,recSize))
  END WrDat;

  PROCEDURE InitPages;

    PROCEDURE Key(    PAGHEAP: CARDINAL;      (*pseudo const*)
                     _PAGHEAP: LONGCARD;
                      size   : LONGCARD;
                      make   : MAKE;
                      ndxPos : LONGCARD;
                  VAR ndxOfs : LONGCARD;
                  VAR n      : LONGCARD);
    VAR
    pag          : ADDRESS;
    key          : KEYITEM;
    pagPos,pagOfs: LONGCARD;
    BEGIN
      Make(ADR(key),ADR(dat),make);

      (*blank page, first item filled*)
      pagPos:=Size(h);
      pagOfs:=_PAGHEAP DIV 2;

      ALLOCATE(pag,PAGHEAP);Fill(pag,PAGHEAP,0);
      SetPag(AddAddr(pag,VAL(CARDINAL,pagOfs)),VAL(CARDINAL,size),key,recPos);
      Seek(h,pagPos);WrBin(h,pag^,PAGHEAP);
      DEALLOCATE(pag,PAGHEAP);

      (*blank index, first item filled*)
      ndxOfs:=NDXHEAP DIV 2;
      n     :=1;

      WrNdx(h,ndxPos,ndxOfs,0,VAL(CARDINAL,size),key,key,pagPos,pagOfs,n)
    END Key;

  BEGIN
    WrDat;

    Key(VAL(CARDINAL,head.pagHeap),head.pagHeap,head.ndx.size,ndxMake,NDXPOS,head.ndx.ofs,head.ndx.n);
    Key(                  SEQHEAP ,     SEQHEAP,head.seq.size,seqMake,SEQPOS,head.seq.ofs,head.seq.n);
    IF head.aux.size#0 THEN
    Key(VAL(CARDINAL,head.pagHeap),head.pagHeap,head.aux.size,auxMake,AUXPOS,head.aux.ofs,head.aux.n)
    END;

    head.empty:=FALSE;
    head.n    :=1;

    Seek(h,0);WrBin(h,head,SIZE(head))
  END InitPages;

  PROCEDURE WrKey(    PAGHEAP: CARDINAL;      (*pseudo const*)
                     _PAGHEAP: LONGCARD;
                      make   : MAKE;
                      ndxPos : LONGCARD;
                  VAR ndx    : KEY): BOOLEAN;
  VAR
  key                 : KEYITEM;
  ndxLoc              : CARDINAL;
  lkey,hkey           : KEYITEM;
  pagPos,pagOfs,pagRec: LONGCARD;

    PROCEDURE SearchNdx(): BOOLEAN;
    VAR
    lptr,hptr,ptr,m: CARDINAL;
    lcmp,hcmp      : CompareType;

      PROCEDURE Keys(loc: CARDINAL);
      BEGIN
        RdNdx(h,ndxPos,ndx.ofs,loc,VAL(CARDINAL,ndx.size),lkey,hkey,pagPos,pagOfs,pagRec);

        lcmp:=Compare(key,lkey,VAL(CARDINAL,ndx.size));
        hcmp:=Compare(key,hkey,VAL(CARDINAL,ndx.size));

        ndxLoc:=loc
      END Keys;

    BEGIN
      (*caso especial item 0*)
      Keys(0);

         IF  lcmp=equal                     THEN RETURN FALSE
      ELSIF  hcmp=equal                     THEN RETURN FALSE
      ELSIF (lcmp=greater) AND (hcmp=lower) THEN RETURN TRUE
      ELSIF  lcmp=lower                     THEN RETURN TRUE
        END;

      (*caso especial item n-1*)
      Keys(VAL(CARDINAL,ndx.n)-1);

         IF  lcmp=equal                     THEN RETURN FALSE
      ELSIF  hcmp=equal                     THEN RETURN FALSE
      ELSIF (lcmp=greater) AND (hcmp=lower) THEN RETURN TRUE
      ELSIF  hcmp=greater                   THEN RETURN TRUE
        END;

      (*situacion general*)
      lptr:=0;
      hptr:=VAL(CARDINAL,ndx.n)-1;
      LOOP
        ptr:=lptr + (hptr-lptr) DIV 2;
        m  :=hptr-lptr+1;

        (*item central*)
        Keys(ptr);

           IF  lcmp=equal              THEN RETURN FALSE
        ELSIF  hcmp=equal              THEN RETURN FALSE
        ELSIF  lcmp=lower              THEN hptr:=ptr
        ELSIF (hcmp=greater) AND (m=2) THEN RETURN TRUE
        ELSIF  hcmp=greater            THEN lptr:=ptr
         ELSE                               RETURN TRUE
          END
      END
    END SearchNdx;

    PROCEDURE ShiftNdx;
    VAR
    n,len: CARDINAL;

      PROCEDURE Underflow(): BOOLEAN;
      BEGIN
        RETURN ndx.ofs < VAL(LONGCARD,len)
      END Underflow;

      PROCEDURE Overflow(): BOOLEAN;
      BEGIN
        RETURN ndx.ofs + (ndx.n+1)*VAL(LONGCARD,len) > NDXHEAP
      END Overflow;

      PROCEDURE Relocate;
      VAR
      mem,source,dest: ADDRESS;
      memHeap,memSize: CARDINAL;
      BEGIN
        memSize:=n*len      ;
        memHeap:=n*len + len;        (*extra space for shift*)

        ALLOCATE(mem,memHeap);
        Seek(h,ndxPos+ndx.ofs);Read(h,mem^,memSize);

        source:=AddAddr(mem,(ndxLoc+1)*len);
        dest  :=AddAddr(mem,(ndxLoc+2)*len);

        Move(source,dest,(n-ndxLoc-1)*len);

        ndx.n  :=ndx.n + 1;
        ndx.ofs:=(NDXHEAP-VAL(LONGCARD,memHeap)) DIV 2;

        Seek(h,ndxPos+ndx.ofs);WrBin(h,mem^,memHeap);

        DEALLOCATE(mem,memHeap)
      END Relocate;

      PROCEDURE Bottom;
      VAR
      mem: ADDRESS;
      BEGIN
        ALLOCATE(mem,len);

        Seek(h,ndxPos+ndx.ofs); Read(h,mem^,len);
        DEC(ndx.ofs,VAL(LONGCARD,len));
        Seek(h,ndxPos+ndx.ofs);WrBin(h,mem^,len);

        DEALLOCATE(mem,len)
      END Bottom;

      PROCEDURE High;
      VAR
      mem    : ADDRESS;
      memSize: CARDINAL;
      BEGIN
        memSize:=(n-ndxLoc-1)*len;

        ALLOCATE(mem,memSize);

        Seek(h,ndxPos+ndx.ofs + (VAL(LONGCARD,ndxLoc)+1)*VAL(LONGCARD,len)); Read(h,mem^,memSize);
        Seek(h,ndxPos+ndx.ofs + (VAL(LONGCARD,ndxLoc)+2)*VAL(LONGCARD,len));WrBin(h,mem^,memSize);

        DEALLOCATE(mem,memSize)
      END High;

      PROCEDURE Low;
      VAR
      memSize: CARDINAL;
      mem    : ADDRESS;
      BEGIN
        memSize:=(ndxLoc+1)*len;

        ALLOCATE(mem,memSize);

        Seek(h,ndxPos+ndx.ofs); Read(h,mem^,memSize);
        DEC(ndx.ofs,VAL(LONGCARD,len));
        Seek(h,ndxPos+ndx.ofs);WrBin(h,mem^,memSize);

        DEALLOCATE(mem,memSize)
      END Low;

    BEGIN
      n  :=VAL(CARDINAL,ndx.n);
      len:=NdxLen(VAL(CARDINAL,ndx.size));

      IF Underflow() OR Overflow() THEN
         Relocate;RETURN
      END;

         IF ndxLoc=n-1            THEN
      ELSIF ndxLoc=0              THEN Bottom
      ELSIF ndxLoc+1 > n-ndxLoc-1 THEN High
       ELSE                            Low
        END;

      INC(ndx.n)
    END ShiftNdx;

    PROCEDURE WrPag(n: CARDINAL): BOOLEAN;
    TYPE
    SEARCH= (before,after,inside,duplicate);

    VAR
    pag,_pag              : ADDRESS;
    pagHeap               : CARDINAL;
    size,len,ndxLen,pagMax: CARDINAL;
    pagLoc                : CARDINAL;

      PROCEDURE Vars;
      BEGIN
        size  :=VAL(CARDINAL,ndx.size);
        ndxLen:=NdxLen(size);
        len   :=PagLen(size);
        pagMax:=PAGHEAP DIV len
      END Vars;

      PROCEDURE RdPag;
      VAR
      pagSize: CARDINAL;
      BEGIN
        pagSize:=      n*len      ;
        pagHeap:=len + n*len + len;  (*extra space for shift*)

        ALLOCATE(pag,pagHeap);
        Seek(h,pagPos+pagOfs);Read(h,AddAddr(pag,len)^,pagSize)
      END RdPag;

      PROCEDURE SearchPag(): SEARCH;
      VAR
      lptr,hptr,ptr,m: CARDINAL;
      cmp            : CompareType;

        PROCEDURE Key(loc: CARDINAL);
        VAR
        _key: KEYITEM;
        _   : LONGCARD;
        BEGIN
          GetPag(AddAddr(pag,(loc+1)*len),size,_key,_);

          cmp:=Compare(key,_key,size);

          pagLoc:=loc
        END Key;

      BEGIN
        (*caso especial item 0*)
        Key(0);

           IF cmp=equal THEN Error("bt/Add/WrKey/WrPag/SearchPag/1")
        ELSIF cmp=lower THEN RETURN before
          END;

        (*caso especial item n-1*)
        Key(n-1);

           IF cmp=equal   THEN Error("bt/Add/WrKey/WrPag/SearchPag/2")
        ELSIF cmp=greater THEN RETURN after
          END;

        (*situacion general*)
        lptr:=0;
        hptr:=n-1;
        LOOP
          ptr:=lptr + (hptr-lptr) DIV 2;
          m  :=hptr-lptr+1;

          (*item central*)
          Key(ptr);

             IF cmp=equal THEN RETURN duplicate
          ELSIF cmp=lower THEN hptr:=ptr
          ELSIF m  =2     THEN RETURN inside
           ELSE                lptr:=ptr
            END
        END
      END SearchPag;

      PROCEDURE Split;
      VAR
      newPag     : ADDRESS;
      n1,n2      : CARDINAL;
      _lkey,_hkey: KEYITEM;
      _          : LONGCARD;
      BEGIN
        n1:=VAL(CARDINAL,pagRec) DIV 2;
        n2:=VAL(CARDINAL,pagRec)-n1;

        (*old page*)
        GetPag(AddAddr(_pag,(n1-1)*len),size,_hkey,_);
        WrNdx(h,ndxPos,ndx.ofs,ndxLoc,size,lkey,_hkey,pagPos,pagOfs,VAL(LONGCARD,n1));

        (*new page*)
        pagPos:=Size(h);
        pagOfs:=(_PAGHEAP-VAL(LONGCARD,n2)*VAL(LONGCARD,len)) DIV 2;

        ALLOCATE(newPag,PAGHEAP);Fill(newPag,PAGHEAP,0);
        Move(AddAddr(_pag,n1*len),AddAddr(newPag,VAL(CARDINAL,pagOfs)),n2*len);
        Seek(h,pagPos);WrBin(h,newPag^,PAGHEAP);
        DEALLOCATE(newPag,PAGHEAP);

        GetPag(AddAddr(_pag,n1*len),size,_lkey,_);
        ShiftNdx;
        WrNdx(h,ndxPos,ndx.ofs,ndxLoc+1,size,_lkey,hkey,pagPos,pagOfs,VAL(LONGCARD,n2))
      END Split;

      PROCEDURE Relocate;
      BEGIN
        pagOfs:=(_PAGHEAP - (VAL(LONGCARD,pagHeap)-VAL(LONGCARD,len))) DIV 2;
        Seek(h,pagPos+pagOfs);WrBin(h,_pag^,pagHeap-len)
      END Relocate;

      PROCEDURE Before;
      BEGIN
        lkey  :=key;
        pagRec:=pagRec+1;

        SetPag(pag,size,key,recPos);_pag:=pag;

        IF pagOfs<VAL(LONGCARD,len) THEN
           Relocate
        ELSE
           DEC(pagOfs,VAL(LONGCARD,len));
           Seek(h,pagPos+pagOfs);WrBin(h,_pag^,len)
        END
      END Before;

      PROCEDURE After;
      VAR
      last: ADDRESS;
      BEGIN
        last  :=AddAddr(pag,(n+1)*len);

        hkey  :=key;
        pagRec:=pagRec+1;

        SetPag(last,size,key,recPos);_pag:=AddAddr(pag,len);

        IF pagOfs + (VAL(LONGCARD,pagHeap)-VAL(LONGCARD,len)) > _PAGHEAP THEN
           Relocate
        ELSE
           Seek(h,pagPos+pagOfs + VAL(LONGCARD,n)*VAL(LONGCARD,len));WrBin(h,last^,len)
        END
      END After;

      PROCEDURE Inside;

        PROCEDURE Underflow(): BOOLEAN;
        BEGIN
          RETURN pagOfs < VAL(LONGCARD,len)
        END Underflow;

        PROCEDURE Overflow(): BOOLEAN;
        BEGIN
          RETURN pagOfs + (VAL(LONGCARD,n)+1)*VAL(LONGCARD,len) > _PAGHEAP
        END Overflow;

        PROCEDURE Relocate;
        VAR
        source,dest: ADDRESS;
        BEGIN
          source:=AddAddr(pag,(pagLoc+2)*len);
          dest  :=AddAddr(pag,(pagLoc+3)*len);

          Move(source,dest,(n-pagLoc-1)*len);
          SetPag(source,size,key,recPos);_pag:=AddAddr(pag,len);

          pagRec:=pagRec+1;
          pagOfs:=(_PAGHEAP - (VAL(LONGCARD,pagHeap)-VAL(LONGCARD,len))) DIV 2;

          Seek(h,pagPos+pagOfs);WrBin(h,_pag^,pagHeap-len)
        END Relocate;

        PROCEDURE High;
        VAR
        source,dest: ADDRESS;
        BEGIN
          source:=AddAddr(pag,(pagLoc+2)*len);
          dest  :=AddAddr(pag,(pagLoc+3)*len);

          Move(source,dest,(n-pagLoc-1)*len);
          SetPag(source,size,key,recPos);_pag:=AddAddr(pag,len);

          pagRec:=pagRec+1;

          Seek(h,pagPos+pagOfs + (VAL(LONGCARD,pagLoc)+1)*VAL(LONGCARD,len));

          WrBin(h,source^,(n-pagLoc)*len)
        END High;

        PROCEDURE Low;
        BEGIN
          pagRec:=pagRec+1;
          pagOfs:=pagOfs-VAL(LONGCARD,len);

          Move(AddAddr(pag,len),pag,(pagLoc+1)*len);
          SetPag(AddAddr(pag,(pagLoc+1)*len),size,key,recPos);_pag:=pag;

          Seek(h,pagPos+pagOfs);WrBin(h,_pag^,(pagLoc+2)*len)
        END Low;

      BEGIN
        IF Underflow() OR Overflow() THEN
           Relocate;RETURN
        END;

        IF pagLoc+1 > n-pagLoc-1 THEN
           High
        ELSE
           Low
        END
      END Inside;

      PROCEDURE FreePag;
      BEGIN
        DEALLOCATE(pag,pagHeap)
      END FreePag;

    BEGIN
      Vars;RdPag;

      CASE SearchPag() OF
         before: Before
        |after : After
        |inside: Inside
      ELSE
                 FreePag;RETURN FALSE
      END;

      WrNdx(h,ndxPos,ndx.ofs,ndxLoc,size,lkey,hkey,pagPos,pagOfs,pagRec);

      IF VAL(CARDINAL,pagRec)>=pagMax-2 THEN
         Split
      END;

      FreePag;RETURN TRUE
    END WrPag;

  BEGIN
    Make(ADR(key),ADR(dat),make);

       IF NOT SearchNdx()                 THEN RETURN FALSE
    ELSIF NOT WrPag(VAL(CARDINAL,pagRec)) THEN RETURN FALSE
      END;

    IF VAL(CARDINAL,ndx.n)>=NdxMax(VAL(CARDINAL,ndx.size))-2 THEN
       head.full:=TRUE
    END;

    RETURN TRUE
  END WrKey;

  PROCEDURE LoadSeq;
  VAR
  bak    : ADDRESS;
  bakHeap: CARDINAL;
  BEGIN
    bakHeap:=SIZE(dat);

    ALLOCATE(bak,bakHeap);
    Move(ADR(dat),bak,bakHeap);

       IF bakHeap<4              THEN Error("bt/Add/LoadSeq/1 - record size underflow")
    ELSIF NOT Find(id,bak^,main) THEN Error("bt/Add/LoadSeq/2 - record not found")
      END;

    Move(bak,ADR(dat),4); (*seq number*)

    DEALLOCATE(bak,bakHeap)
  END LoadSeq;

BEGIN
  (*get item from list*)
  file   :=File(id);
  h      :=file^.h;
  recSize:=file^.recSize;
  ndxMake:=file^.ndxMake;
  auxMake:=file^.auxMake;

  (*read head*)
  Lock(h);Seek(h,0);Read(h,head,SIZE(head));

  (*file room*)
     IF head.empty THEN InitPages;Unlock(h);RETURN TRUE
  ELSIF head.full  THEN Error("bt/Add/1 - index heap full")
    END;

  WrDat;

     IF NOT WrKey(VAL(CARDINAL,head.pagHeap),head.pagHeap,ndxMake,NDXPOS,head.ndx) THEN Unlock(h);LoadSeq;RETURN FALSE
  ELSIF NOT WrKey(                  SEQHEAP ,     SEQHEAP,seqMake,SEQPOS,head.seq) THEN Error("bt/Add/2 - duplicated key (seq)")
  ELSIF head.aux.size=0                                                        THEN
  ELSIF NOT WrKey(VAL(CARDINAL,head.pagHeap),head.pagHeap,auxMake,AUXPOS,head.aux) THEN Error("bt/Add/3 - duplicated key (aux)")
    END;

  (*update head*)
  INC(head.n);

  IF head.n>=16777216 THEN (*2^24 --> seq keysize=24 bits*)
     head.full:=TRUE
  END;

  Seek(h,0);WrBin(h,head,SIZE(head));

  Unlock(h);RETURN TRUE
END Add;

PROCEDURE Change(id : LONGCARD;
                 dat: ARRAY OF BYTE): BOOLEAN;
VAR
_h    : FIO.File;
_aux  : BOOLEAN;
recPos: LONGCARD;

  PROCEDURE Find(index: IndexType): BOOLEAN;
  VAR
  h        : FIO.File;
  q        : QUERY;
  key      : KEYITEM;
  pos,ofs,n: LONGCARD;
  ndxLoc   : CARDINAL;
  r        : BOOLEAN;

    PROCEDURE WrDat;
    BEGIN
      recPos:=Size(h);Seek(h,recPos);WrBin(h,dat,q.recSize)
    END WrDat;

    PROCEDURE FoundBottom(): BOOLEAN;
    VAR
    pag: ADDRESS;
    BEGIN
      ALLOCATE(pag,q.pagLen);

      Seek(h, pos+ofs); Read(h,pag^,q.pagLen);xSetPag(pag,q.ndxSize,recPos);
      Seek(h, pos+ofs);WrBin(h,pag^,q.pagLen);

      DEALLOCATE(pag,q.pagLen);RETURN TRUE
    END FoundBottom;

    PROCEDURE FoundTop(): BOOLEAN;
    VAR
    pag: ADDRESS;
    BEGIN
      ALLOCATE(pag,q.pagLen);

      Seek(h, pos+ofs + (n-1)*VAL(LONGCARD,q.pagLen)); Read(h,pag^,q.pagLen);xSetPag(pag,q.ndxSize,recPos);
      Seek(h, pos+ofs + (n-1)*VAL(LONGCARD,q.pagLen));WrBin(h,pag^,q.pagLen);

      DEALLOCATE(pag,q.pagLen);RETURN TRUE
    END FoundTop;

    PROCEDURE FoundThis(): BOOLEAN;
    VAR
    pag        : ADDRESS;
    pagHeap,loc: CARDINAL;
    _          : LONGCARD;
    BEGIN
      pagHeap:=VAL(CARDINAL,n)*q.pagLen;
      ALLOCATE(pag,pagHeap);

      Seek(h,pos+ofs);Read(h,pag^,pagHeap);

      IF NOT QSearchPag(q,pag,key,q.ndxSize,VAL(CARDINAL,n),_,loc) THEN
         DEALLOCATE(pag,pagHeap);RETURN FALSE
      END;

      xSetPag(AddAddr(pag,loc*q.pagLen),q.ndxSize,recPos);

      Seek(h,pos+ofs + VAL(LONGCARD,loc)*VAL(LONGCARD,q.pagLen));
      WrBin(h,AddAddr(pag,loc*q.pagLen)^,q.pagLen);

      DEALLOCATE(pag,pagHeap);RETURN TRUE
    END FoundThis;

  BEGIN
    IF NOT QCommon(id,h,q,index,FALSE,index=main) THEN
       RETURN FALSE
    END;

    IF index=main THEN
       WrDat;

      _h  :=h;
      _aux:=q.aux
    END;

    Make(ADR(key),ADR(dat),q.make);

    CASE QSearch(q,key,pos,ofs,n,ndxLoc,TRUE) OF
       foundBottom  : r:=FoundBottom()
      |foundTop     : r:=FoundTop()
      |foundThis    : r:=FoundThis()
      |notFoundFirst: r:=FALSE
      |notFoundLast : r:=FALSE
      |notFound     : r:=FALSE
    ELSE
                      Error("bt/Change/Find/1")
    END;

    IF (index=main) AND (NOT r) THEN
       Unlock(h)
    END;

    RETURN r
  END Find;

BEGIN
     IF NOT Find(main) THEN RETURN FALSE
  ELSIF NOT Find(seq ) THEN Unlock(_h);Error("bt/Change/1")
  ELSIF NOT _aux       THEN Unlock(_h);RETURN TRUE
  ELSIF NOT Find(aux ) THEN Unlock(_h);Error("bt/Change/2")
    END;

  Unlock(_h);RETURN TRUE
END Change;

PROCEDURE Query(    id   : LONGCARD;
                VAR q    : QUERY;
                    index: IndexType);
VAR
h: FIO.File;
BEGIN
  IF QCommon(id,h,q,index,TRUE,TRUE) THEN
     QSort(q,q.ndx,1,q.n);
     QRead(q);

     Unlock(q.h)
  END
END Query;

PROCEDURE Query2(    id   : LONGCARD;
                 VAR q    : QUERY;
                     dat  : ARRAY OF BYTE;
                     len  : CARDINAL;
                     index: IndexType);
TYPE
WALK= (this,up,down);

VAR
h                        : FIO.File;
key                      : KEYITEM;
pag                      : ADDRESS;
ndxLoc,bottom,top,pagHeap: CARDINAL;
pagPos,pagOfs,pagN       : LONGCARD;

  PROCEDURE RdPag;
  BEGIN
    pagHeap:=VAL(CARDINAL,pagN)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pagPos+pagOfs);Read(h,pag^,pagHeap)
  END RdPag;

  PROCEDURE Bottom(): BOOLEAN;

    PROCEDURE FoundThis(): BOOLEAN;
    VAR
    r : BOOLEAN;
    _ : LONGCARD;
    __: CARDINAL;
    BEGIN
      bottom:=ndxLoc;

      RdPag;
      r:=QSearchPag(q,pag,key,len,VAL(CARDINAL,pagN),_,__);
      DEALLOCATE(pag,pagHeap);

      RETURN r
    END FoundThis;

    PROCEDURE NotFoundFirst(): BOOLEAN;
    VAR
    pos,ofs,n: LONGCARD;
    lkey,hkey: KEYITEM;
    BEGIN
      bottom:=0;

      GetNdx(q.ndx,q.ndxSize,lkey,hkey,pos,ofs,n);

      RETURN Compare(key,lkey,len)=equal
    END NotFoundFirst;

    PROCEDURE NotFound(): BOOLEAN;
    VAR
    pos,ofs,n: LONGCARD;
    lkey,hkey: KEYITEM;
    BEGIN
       bottom:=ndxLoc+1;

      GetNdx(AddAddr(q.ndx,bottom*q.ndxLen),q.ndxSize,lkey,hkey,pos,ofs,n);

      RETURN Compare(key,lkey,len)=equal
    END NotFound;

  BEGIN
    Make(ADR(key),ADR(dat),q.make);PadKey(key,len,0);

    CASE QSearch(q,key,pagPos,pagOfs,pagN,ndxLoc,FALSE) OF
       foundBottom  : bottom:=ndxLoc;RETURN TRUE
      |foundTop     : bottom:=ndxLoc;RETURN TRUE
      |foundThis    :                RETURN FoundThis()
      |notFoundFirst:                RETURN NotFoundFirst()
      |notFoundLast :                RETURN FALSE
      |notFound     :                RETURN NotFound()
    ELSE
                      Error("bt/Query2/Bottom/1")
    END
  END Bottom;

  PROCEDURE Top(): BOOLEAN;

    PROCEDURE FoundThis(): BOOLEAN;
    VAR
    r : BOOLEAN;
    _ : LONGCARD;
    __: CARDINAL;
    BEGIN
      top:=ndxLoc;

      RdPag;
      r:=QSearchPag(q,pag,key,len,VAL(CARDINAL,pagN),_,__);
      DEALLOCATE(pag,pagHeap);

      RETURN r
    END FoundThis;

    PROCEDURE NotFoundLast(): BOOLEAN;
    VAR
    pos,ofs,n: LONGCARD;
    lkey,hkey: KEYITEM;
    BEGIN
      top:=ndxLoc;

      GetNdx(AddAddr(q.ndx,top*q.ndxLen),q.ndxSize,lkey,hkey,pos,ofs,n);

      RETURN Compare(key,hkey,len)=equal
    END NotFoundLast;

    PROCEDURE NotFound(): BOOLEAN;
    VAR
    pos,ofs,n: LONGCARD;
    lkey,hkey: KEYITEM;
    BEGIN
      top:=ndxLoc;

      GetNdx(AddAddr(q.ndx,top*q.ndxLen),q.ndxSize,lkey,hkey,pos,ofs,n);

      RETURN Compare(key,hkey,len)=equal
    END NotFound;

  BEGIN
    PadKey(key,len,0FFH);

    CASE QSearch(q,key,pagPos,pagOfs,pagN,ndxLoc,FALSE) OF
       foundBottom  : top:=ndxLoc;RETURN TRUE
      |foundTop     : top:=ndxLoc;RETURN TRUE
      |foundThis    :             RETURN FoundThis()
      |notFoundFirst:             RETURN FALSE
      |notFoundLast :             RETURN NotFoundLast()
      |notFound     :             RETURN NotFound()
    ELSE
                      Error("bt/Query2/Top/1")
    END
  END Top;

  PROCEDURE PagBlock(    i         : ADDRESS;
                         n         : CARDINAL;
                     VAR first,last: CARDINAL): BOOLEAN;
  VAR
  inBlock          : BOOLEAN;
  _key             : KEYITEM;
  _                : LONGCARD;
  k                : CARDINAL;
  cmp,item_1,item_n: CompareType;
  BEGIN
    (*bordes*)
    GetPag(        i                ,q.ndxSize,_key,_);item_1:=Compare(key,_key,len);
    GetPag(AddAddr(i,(n-1)*q.pagLen),q.ndxSize,_key,_);item_n:=Compare(key,_key,len);

       IF  item_1=lower                       THEN                    RETURN FALSE
    ELSIF  item_n=greater                     THEN                    RETURN FALSE
    ELSIF (item_1=equal  ) AND (item_n=equal) THEN first:=0;last:=n-1;RETURN TRUE
      END;

    (*barrido*)
    inBlock:=FALSE;first:=0;last:=0;
    FOR k:=1 TO n DO
      GetPag(i,q.ndxSize,_key,_);cmp:=Compare(key,_key,len);

         IF inBlock AND (cmp#equal) THEN RETURN TRUE
      ELSIF inBlock                 THEN last:=k-1
      ELSIF              cmp=equal  THEN first:=k-1;last:=k-1;inBlock:=TRUE
        END;

      IncAddr(i,q.pagLen)
    END;

    RETURN inBlock
  END PagBlock;

  PROCEDURE This(ndxLoc: CARDINAL;
                 walk  : WALK);
  VAR
  pag,i             : ADDRESS;
  pagHeap,first,last: CARDINAL;
  lkey,hkey         : KEYITEM;
  pos,ofs,n         : LONGCARD;
  BEGIN
    i:=AddAddr(q.ndx,ndxLoc*q.ndxLen);

    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    IF NOT PagBlock(pag,VAL(CARDINAL,n),first,last) THEN
       Error("bt/Query2/This/1")
    END;

    QPages(q,pag,i,pagHeap,n,first,last);

    CASE walk OF
       this: q.n:=ndxLoc+1;q.K:=ndxLoc;q.k:=ndxLoc
      |up  : q.n:=ndxLoc+1
      |down:               q.K:=ndxLoc;q.k:=ndxLoc
    ELSE
             Error("bt/Query2/This/2")
    END
  END This;

  PROCEDURE Pages(): CARDINAL;
  BEGIN
    RETURN top-bottom+1
  END Pages;

BEGIN
     IF NOT QCommon(id,h,q,index,TRUE,TRUE) THEN                                                                   RETURN
  ELSIF NOT Bottom()                        THEN CloseQuery(q);                                        Unlock(q.h);RETURN
  ELSIF NOT Top()                           THEN Error("bt/Query2/1")
  ELSIF Pages()=1                           THEN This(bottom,this);                                    Unlock(q.h);RETURN
  ELSIF Pages()=2                           THEN This(bottom,this);This(bottom+1,up);                  Unlock(q.h);RETURN
  ELSIF Pages()=3                           THEN This(bottom,this);This(bottom+1,up);This(bottom+2,up);Unlock(q.h);RETURN
    END;

  QSort(q,AddAddr(q.ndx,(bottom+1)*q.ndxLen),1,Pages()-2);
  QRead(q);

  This(top   ,this);
  This(bottom,down);

  Unlock(q.h)
END Query2;

PROCEDURE Query3(    id   : LONGCARD;
                 VAR q    : QUERY;
                     dat1 : ARRAY OF BYTE;
                     dat2 : ARRAY OF BYTE;
                     len  : CARDINAL;
                     index: IndexType);
VAR
h                 : FIO.File;
key1,key2         : KEYITEM;
ndxLoc,bottom,top : CARDINAL;
pagPos,pagOfs,pagN: LONGCARD;

  PROCEDURE Range(): BOOLEAN;
  VAR
  key1,key2: KEYITEM;
  BEGIN
    Make(ADR(key1),ADR(dat1),q.make);PadKey(key1,len,0   );
    Make(ADR(key2),ADR(dat2),q.make);PadKey(key2,len,0FFH);

    RETURN Compare(key1,key2,len)#greater
  END Range;

  PROCEDURE Bottom(): BOOLEAN;
  BEGIN
    Make(ADR(key1),ADR(dat1),q.make);PadKey(key1,len,0);

    CASE QSearch(q,key1,pagPos,pagOfs,pagN,ndxLoc,FALSE) OF
       foundBottom  : bottom:=ndxLoc  ;RETURN TRUE
      |foundTop     : bottom:=ndxLoc  ;RETURN TRUE
      |foundThis    : bottom:=ndxLoc  ;RETURN TRUE
      |notFoundFirst: bottom:=ndxLoc  ;RETURN TRUE
      |notFoundLast :                  RETURN FALSE
      |notFound     : bottom:=ndxLoc+1;RETURN TRUE
    ELSE
                      Error("bt/Query3/Bottom/1")
    END
  END Bottom;

  PROCEDURE Top(): BOOLEAN;
  BEGIN
    Make(ADR(key2),ADR(dat2),q.make);PadKey(key2,len,0FFH);

    CASE QSearch(q,key2,pagPos,pagOfs,pagN,ndxLoc,FALSE) OF
       foundBottom  : top:=ndxLoc;RETURN TRUE
      |foundTop     : top:=ndxLoc;RETURN TRUE
      |foundThis    : top:=ndxLoc;RETURN TRUE
      |notFoundFirst:             RETURN FALSE
      |notFoundLast : top:=ndxLoc;RETURN TRUE
      |notFound     : top:=ndxLoc;RETURN TRUE
    ELSE
                      Error("bt/Query3/Top/1")
    END
  END Top;

  PROCEDURE Pages(): CARDINAL;
  BEGIN
    RETURN top-bottom+1
  END Pages;

  PROCEDURE LowLocation(i: ADDRESS;
                        n: CARDINAL): CARDINAL;
  VAR
  _key: KEYITEM;
  _   : LONGCARD;
  k   : CARDINAL;
  BEGIN
    FOR k:=1 TO n DO
      GetPag(i,q.ndxSize,_key,_);

      IF Compare(key1,_key,len)#greater THEN
         RETURN k-1
      END;

      IncAddr(i,q.pagLen)
    END;

    Error("bt/Query3/LowLocation/1")
  END LowLocation;

  PROCEDURE HighLocation(i: ADDRESS;
                         n: CARDINAL): CARDINAL;
  VAR
  _key: KEYITEM;
  _   : LONGCARD;
  k   : CARDINAL;
  BEGIN
    IncAddr(i,(n-1)*q.pagLen);
    FOR k:=n TO 1 BY -1 DO
      GetPag(i,q.ndxSize,_key,_);

      IF Compare(key2,_key,len)#lower THEN
         RETURN k-1
      END;

      DecAddr(i,q.pagLen)
    END;

    Error("bt/Query3/HighLocation/1")
  END HighLocation;

  PROCEDURE Overlap(ndxLoc: CARDINAL): BOOLEAN;
  VAR
  pag      : ADDRESS;
  pagHeap  : CARDINAL;
  lkey,hkey: KEYITEM;
  pos,ofs,n: LONGCARD;
  r        : BOOLEAN;
  BEGIN
    GetNdx(AddAddr(q.ndx,ndxLoc*q.ndxLen),q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    r:=LowLocation(pag,VAL(CARDINAL,n))>HighLocation(pag,VAL(CARDINAL,n));

    DEALLOCATE(pag,pagHeap);RETURN r
  END Overlap;

  PROCEDURE This(ndxLoc: CARDINAL);
  VAR
  pag,i    : ADDRESS;
  pagHeap  : CARDINAL;
  lkey,hkey: KEYITEM;
  pos,ofs,n: LONGCARD;
  BEGIN
    i:=AddAddr(q.ndx,ndxLoc*q.ndxLen);

    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    xSetNdx(i,q.ndxSize,FALSE,pag,VAL(LONGCARD,pagHeap),n,0);

    INC(q.recs,n)
  END This;

  PROCEDURE ThisRange(ndxLoc: CARDINAL);
  VAR
  pag,i    : ADDRESS;
  pagHeap  : CARDINAL;
  lkey,hkey: KEYITEM;
  pos,ofs,n: LONGCARD;
  BEGIN
    i:=AddAddr(q.ndx,ndxLoc*q.ndxLen);

    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    QPages(q,pag,i,pagHeap,n,LowLocation(pag,VAL(CARDINAL,n)),HighLocation(pag,VAL(CARDINAL,n)));

    q.n:=ndxLoc+1;q.K:=ndxLoc;q.k:=ndxLoc
  END ThisRange;

  PROCEDURE ThisDown(ndxLoc: CARDINAL);
  VAR
  pag,i    : ADDRESS;
  pagHeap  : CARDINAL;
  lkey,hkey: KEYITEM;
  pos,ofs,n: LONGCARD;
  BEGIN
    i:=AddAddr(q.ndx,ndxLoc*q.ndxLen);

    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    QPages(q,pag,i,pagHeap,n,LowLocation(pag,VAL(CARDINAL,n)),VAL(CARDINAL,n)-1);

    q.K:=ndxLoc;q.k:=ndxLoc
  END ThisDown;

  PROCEDURE ThisUp(ndxLoc: CARDINAL);
  VAR
  pag,i    : ADDRESS;
  pagHeap  : CARDINAL;
  lkey,hkey: KEYITEM;
  pos,ofs,n: LONGCARD;
  BEGIN
    i:=AddAddr(q.ndx,ndxLoc*q.ndxLen);

    GetNdx(i,q.ndxSize,lkey,hkey,pos,ofs,n);

    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    QPages(q,pag,i,pagHeap,n,0,HighLocation(pag,VAL(CARDINAL,n)));

    q.n:=ndxLoc+1
  END ThisUp;

BEGIN
     IF  NOT QCommon(id,h,q,index,TRUE,TRUE) THEN                                                                 RETURN
  ELSIF  NOT Range()                         THEN CloseQuery(q);                                      Unlock(q.h);RETURN
  ELSIF  NOT Bottom()                        THEN CloseQuery(q);                                      Unlock(q.h);RETURN
  ELSIF  NOT Top()                           THEN CloseQuery(q);                                      Unlock(q.h);RETURN
  ELSIF  top<bottom                          THEN CloseQuery(q);                                      Unlock(q.h);RETURN
  ELSIF (Pages()=1) AND Overlap(bottom)      THEN CloseQuery(q);                                      Unlock(q.h);RETURN
  ELSIF  Pages()=1                           THEN ThisRange(bottom);                                  Unlock(q.h);RETURN
  ELSIF  Pages()=2                           THEN ThisDown (bottom);ThisUp(bottom+1);                 Unlock(q.h);RETURN
  ELSIF  Pages()=3                           THEN ThisDown (bottom);This  (bottom+1);ThisUp(bottom+2);Unlock(q.h);RETURN
    END;

  QSort(q,AddAddr(q.ndx,(bottom+1)*q.ndxLen),1,Pages()-2);
  QRead(q);

  ThisDown(bottom);
  ThisUp  (top   );

  Unlock(q.h)
END Query3;

PROCEDURE CloseQuery(VAR q: QUERY);

  PROCEDURE FreeNdx;
  VAR
  pag          : ADDRESS;
  loaded       : BOOLEAN;
  pagHeap,k,m,_: CARDINAL;

    PROCEDURE FreePag;
    VAR
    rec,i: ADDRESS;
    j    : CARDINAL;
    BEGIN
      i:=pag;
      FOR j:=1 TO m DO
        xGetPag(i,q.ndxSize,rec);

        IF rec#NIL THEN
           DEALLOCATE(rec,q.recSize)
        END;

        IncAddr(i,q.pagLen)
      END
    END FreePag;

  BEGIN
    IF q.recs=0 THEN
       RETURN
    END;

    FOR k:=q.K TO q.n-1 DO
      xGetNdx(AddAddr(q.ndx,k*q.ndxLen),q.ndxSize,loaded,pag,pagHeap,m,_);

      IF pag#NIL THEN
         IF loaded THEN
            FreePag
         END;

         DEALLOCATE(pag,pagHeap)
      END
    END
  END FreeNdx;

BEGIN
  IF q.ndx#NIL THEN
     FreeNdx;DEALLOCATE(q.ndx,q.ndxHeap);q.file^.query:=NIL
  END
END CloseQuery;

PROCEDURE Next(VAR q  : QUERY;
               VAR rec: ARRAY OF BYTE): BOOLEAN;
VAR
loaded     : BOOLEAN;
pag,dat    : ADDRESS;
pagHeap,m,j: CARDINAL;
page       : PAGE;

  PROCEDURE Sort;
  VAR
  pag,i       : ADDRESS;
  j,k,n,__,___: CARDINAL;
  pos         : LONGCARD;
  loaded,_    : BOOLEAN;
  ____        : KEYITEM;
  BEGIN
    ResetIdx;loaded:=FALSE;
    FOR k:=q.k TO q.n-1 DO
      xGetNdx(AddAddr(q.ndx,k*q.ndxLen),q.ndxSize,_,pag,__,n,___);

         IF idxNum+n<=IDXMAX THEN
      ELSIF NOT loaded       THEN Error("bt/Next/Sort/1")
       ELSE                       RETURN
        END;

      i:=pag;
      FOR j:=0 TO n-1 DO
        GetPag(i,q.ndxSize,____,pos);AddIdx(i,pos);
        IncAddr(i,q.pagLen)
      END;

      loaded:=TRUE;

      Move(ADR(loaded),AddAddr(q.ndx,k*q.ndxLen + 2*q.ndxSize),1)
    END
  END Sort;

  PROCEDURE RdRecs;
  VAR
  rec,i,link: ADDRESS;
  k         : CARDINAL;
  pos       : LONGCARD;
  BEGIN
    i:=AddAddr(idx,idxOfs);
    FOR k:=1 TO idxNum DO
      GetIdx(i,link,pos);

      ALLOCATE(rec,q.recSize);
      IF rec=NIL THEN
         Error("bt/Next/RdRecs/1 - not enough memory")
      END;

      Seek(q.h,pos);Read(q.h,rec^,q.recSize);

      Move(ADR(rec),AddAddr(link,q.ndxSize),4);

      IncAddr(i,IDXLEN)
    END
  END RdRecs;

BEGIN
     IF q.recs =0   THEN RETURN FALSE
  ELSIF q.ndx  =NIL THEN RETURN FALSE
  ELSIF q.k   >=q.n THEN RETURN FALSE
    END;

  xGetNdx(AddAddr(q.ndx,q.k*q.ndxLen),q.ndxSize,loaded,pag,pagHeap,m,j);

  IF NOT loaded THEN
     Sort;RdRecs
  END;

  xGetPag(AddAddr(pag,j*q.pagLen),q.ndxSize,dat);

  Move(dat,ADR(rec),q.recSize);
  DEALLOCATE(dat,q.recSize);
  Move(ADR(dat),AddAddr(pag,j*q.pagLen + q.ndxSize),4);

  IF j=m-1 THEN
     INC(q.k);RETURN TRUE
  ELSE
     ZipPage(page,VAL(LONGCARD,pagHeap),VAL(LONGCARD,m),VAL(LONGCARD,j)+1);
     Move(ADR(page),AddAddr(q.ndx,q.k*q.ndxLen + 2*q.ndxSize+1+4),SIZE(PAGE));

     RETURN TRUE
  END
END Next;

PROCEDURE Find(    id   : LONGCARD;
               VAR dat  : ARRAY OF BYTE;
                   index: IndexType): BOOLEAN;
VAR
h        : FIO.File;
q        : QUERY;
key      : KEYITEM;
pos,ofs,n: LONGCARD;
ndxLoc   : CARDINAL;
r        : BOOLEAN;

  PROCEDURE FoundBottom(): BOOLEAN;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h, pos+ofs);Read(h,pag^,q.pagLen );GetPag(pag,q.ndxSize,_key,_pos);
    Seek(h,_pos    );Read(h,dat ,q.recSize);

    DEALLOCATE(pag,q.pagLen);RETURN TRUE
  END FoundBottom;

  PROCEDURE FoundTop(): BOOLEAN;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h, pos+ofs + (n-1)*VAL(LONGCARD,q.pagLen));Read(h,pag^,q.pagLen );GetPag(pag,q.ndxSize,_key,_pos);
    Seek(h,_pos                                   );Read(h,dat ,q.recSize);

    DEALLOCATE(pag,q.pagLen);RETURN TRUE
  END FoundTop;

  PROCEDURE FoundThis(): BOOLEAN;
  VAR
  pag      : ADDRESS;
  pagHeap,_: CARDINAL;
  _pos     : LONGCARD;
  BEGIN
    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    IF NOT QSearchPag(q,pag,key,q.ndxSize,VAL(CARDINAL,n),_pos,_) THEN
       DEALLOCATE(pag,pagHeap);RETURN FALSE
    END;

    Seek(h,_pos);Read(h,dat,q.recSize);

    DEALLOCATE(pag,pagHeap);RETURN TRUE
  END FoundThis;

BEGIN
  IF NOT QCommon(id,h,q,index,FALSE,TRUE) THEN
     RETURN FALSE
  END;

  Make(ADR(key),ADR(dat),q.make);

  CASE QSearch(q,key,pos,ofs,n,ndxLoc,TRUE) OF
     foundBottom  : r:=FoundBottom()
    |foundTop     : r:=FoundTop()
    |foundThis    : r:=FoundThis()
    |notFoundFirst: r:=FALSE
    |notFoundLast : r:=FALSE
    |notFound     : r:=FALSE
  ELSE
                    Error("bt/Find/1")
  END;

  Unlock(q.h);RETURN r
END Find;

PROCEDURE Delete(id : LONGCARD;
                 dat: ARRAY OF BYTE);
VAR
h            : FIO.File;
q            : QUERY;
index        : IndexType;
key          : KEYITEM;
pos,ofs,n    : LONGCARD;
ndxLoc,pagLoc: CARDINAL;
search       : QSEARCH;
found        : BOOLEAN;

  PROCEDURE FoundBottom;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h, pos+ofs);Read(h,pag^,q.pagLen );GetPag(pag,q.ndxSize,_key,_pos);
    Seek(h,_pos    );Read(h,dat ,q.recSize);

    DEALLOCATE(pag,q.pagLen)
  END FoundBottom;

  PROCEDURE FoundTop;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h, pos+ofs + (n-1)*VAL(LONGCARD,q.pagLen));Read(h,pag^,q.pagLen );GetPag(pag,q.ndxSize,_key,_pos);
    Seek(h,_pos                                   );Read(h,dat ,q.recSize);

    DEALLOCATE(pag,q.pagLen)
  END FoundTop;

  PROCEDURE FoundThis;
  VAR
  pag    : ADDRESS;
  pagHeap: CARDINAL;
  _pos   : LONGCARD;
  BEGIN
    pagHeap:=VAL(CARDINAL,n)*q.pagLen;
    ALLOCATE(pag,pagHeap);

    Seek(h,pos+ofs);Read(h,pag^,pagHeap);

    IF NOT QSearchPag(q,pag,key,q.ndxSize,VAL(CARDINAL,n),_pos,pagLoc) THEN
       DEALLOCATE(pag,pagHeap);found:=FALSE;RETURN
    END;

    Seek(h,_pos);Read(h,dat,q.recSize);

    DEALLOCATE(pag,pagHeap);found:=TRUE
  END FoundThis;

  PROCEDURE Reset;
  VAR
  head: HEAD;
  BEGIN
    Seek(h,0);Read(h,head,SIZE(head));

    head.empty  :=TRUE;
    head.full   :=FALSE;
    head.n      :=0;

    head.seq.ofs:=0;
    head.seq.n  :=0;

    head.ndx.ofs:=0;
    head.ndx.n  :=0;

    head.aux.ofs:=0;
    head.aux.n  :=0;

    Seek(h,0);WrBin(h,head,SIZE(head))
  END Reset;

  PROCEDURE RemoveNdx;
  VAR
  head: HEAD;

    PROCEDURE DecN;
    BEGIN
      CASE index OF
         seq : DEC(head.seq.n)
        |main: DEC(head.ndx.n)
        |aux : DEC(head.aux.n)
      ELSE
               Error("bt/Delete/RemoveNdx/DecN/1")
      END
    END DecN;

    PROCEDURE IncOfs;
    BEGIN
      CASE index OF
         seq : INC(head.seq.ofs,VAL(LONGCARD,q.ndxLen))
        |main: INC(head.ndx.ofs,VAL(LONGCARD,q.ndxLen))
        |aux : INC(head.aux.ofs,VAL(LONGCARD,q.ndxLen))
      ELSE
               Error("bt/Delete/RemoveNdx/IncOfs/1")
      END
    END IncOfs;

    PROCEDURE Low;
    VAR
    ndx    : ADDRESS;
    ndxHeap: CARDINAL;
    BEGIN
      ndxHeap:=ndxLoc*q.ndxLen;

      ALLOCATE(ndx,ndxHeap);

      Seek(h,q.ndxPos+q.ndxOfs                         ); Read(h,ndx^,ndxHeap);
      Seek(h,q.ndxPos+q.ndxOfs + VAL(LONGCARD,q.ndxLen));WrBin(h,ndx^,ndxHeap);

      DEALLOCATE(ndx,ndxHeap)
    END Low;

    PROCEDURE High;
    VAR
    ndx    : ADDRESS;
    ndxHeap: CARDINAL;
    BEGIN
      ndxHeap:=(q.n-ndxLoc-1)*q.ndxLen;

      ALLOCATE(ndx,ndxHeap);

      Seek(h,q.ndxPos+q.ndxOfs + (VAL(LONGCARD,ndxLoc)+1)*VAL(LONGCARD,q.ndxLen)); Read(h,ndx^,ndxHeap);
      Seek(h,q.ndxPos+q.ndxOfs + (VAL(LONGCARD,ndxLoc)  )*VAL(LONGCARD,q.ndxLen));WrBin(h,ndx^,ndxHeap);

      DEALLOCATE(ndx,ndxHeap)
    END High;

  BEGIN
    Seek(h,0);Read(h,head,SIZE(HEAD));

       IF ndxLoc =0         THEN      DecN;IncOfs
    ELSIF ndxLoc =q.n-1     THEN      DecN
    ELSIF ndxLoc>=q.n DIV 2 THEN High;DecN
     ELSE                        Low ;DecN;IncOfs
      END;

    Seek(h,0);WrBin(h,head,SIZE(HEAD))
  END RemoveNdx;

  PROCEDURE RemoveBottom;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h,pos+ofs + VAL(LONGCARD,q.pagLen));Read(h,pag^,q.pagLen);
    GetPag(pag,q.ndxSize,_key,_pos);

    WrNdx(h,q.ndxPos,q.ndxOfs,ndxLoc,q.ndxSize,_key,_hKey,pos,ofs+VAL(LONGCARD,q.pagLen),n-1);

    DEALLOCATE(pag,q.pagLen)
  END RemoveBottom;

  PROCEDURE RemoveTop;
  VAR
  pag : ADDRESS;
  _key: KEYITEM;
  _pos: LONGCARD;
  BEGIN
    ALLOCATE(pag,q.pagLen);

    Seek(h,pos+ofs + (n-2)*VAL(LONGCARD,q.pagLen));Read(h,pag^,q.pagLen);
    GetPag(pag,q.ndxSize,_key,_pos);

    WrNdx(h,q.ndxPos,q.ndxOfs,ndxLoc,q.ndxSize,_lKey,_key,pos,ofs,n-1);

    DEALLOCATE(pag,q.pagLen)
  END RemoveTop;

  PROCEDURE RemoveThis;

    PROCEDURE Low;
    VAR
    pag    : ADDRESS;
    pagHeap: CARDINAL;
    BEGIN
      pagHeap:=pagLoc*q.pagLen;

      ALLOCATE(pag,pagHeap);

      Seek(h,pos+ofs                         ); Read(h,pag^,pagHeap);
      Seek(h,pos+ofs + VAL(LONGCARD,q.pagLen));WrBin(h,pag^,pagHeap);

      WrNdx(h,q.ndxPos,q.ndxOfs,ndxLoc,q.ndxSize,_lKey,_hKey,pos,ofs+VAL(LONGCARD,q.pagLen),n-1);

      DEALLOCATE(pag,pagHeap);
    END Low;

    PROCEDURE High;
    VAR
    pag    : ADDRESS;
    pagHeap: CARDINAL;
    BEGIN
      pagHeap:=(VAL(CARDINAL,n)-pagLoc-1)*q.pagLen;

      ALLOCATE(pag,pagHeap);

      Seek(h,pos+ofs + (VAL(LONGCARD,pagLoc)+1)*VAL(LONGCARD,q.pagLen)); Read(h,pag^,pagHeap);
      Seek(h,pos+ofs + (VAL(LONGCARD,pagLoc)  )*VAL(LONGCARD,q.pagLen));WrBin(h,pag^,pagHeap);

      WrNdx(h,q.ndxPos,q.ndxOfs,ndxLoc,q.ndxSize,_lKey,_hKey,pos,ofs,n-1);

      DEALLOCATE(pag,pagHeap)
    END High;

  BEGIN
       IF pagLoc =0                     THEN Error("bt/Delete/RemoveThis/1")
    ELSIF pagLoc =VAL(CARDINAL,n)-1     THEN Error("bt/Delete/RemoveThis/2")
    ELSIF pagLoc>=VAL(CARDINAL,n) DIV 2 THEN High
     ELSE                                    Low
      END
  END RemoveThis;

  PROCEDURE Combine;
  TYPE
  CELL= RECORD
          lkey,hkey: KEYITEM;
          pos,ofs,n: LONGCARD;
          ndxLoc   : CARDINAL
        END;

  VAR
  lcell,hcell,cell1,cell2,cell3: CELL;

    PROCEDURE RdCell2(    ndxLoc     : CARDINAL;
                      VAR cell1,cell2: CELL);
    VAR
    cell,i  : ADDRESS;
    cellHeap: CARDINAL;
    BEGIN
      cellHeap:=2*q.ndxLen;
      ALLOCATE(cell,cellHeap);

      Seek(h,q.ndxPos+q.ndxOfs + VAL(LONGCARD,ndxLoc)*VAL(LONGCARD,q.ndxLen));Read(h,cell^,cellHeap);

      i:=cell;
      GetNdx(i,q.ndxSize,cell1.lkey,cell1.hkey,cell1.pos,cell1.ofs,cell1.n);IncAddr(i,q.ndxLen);
      GetNdx(i,q.ndxSize,cell2.lkey,cell2.hkey,cell2.pos,cell2.ofs,cell2.n);

      cell1.ndxLoc:=ndxLoc;
      cell2.ndxLoc:=ndxLoc+1;

      DEALLOCATE(cell,cellHeap)
    END RdCell2;

    PROCEDURE RdCell3(    ndxLoc           : CARDINAL;
                      VAR cell1,cell2,cell3: CELL);
    VAR
    cell,i  : ADDRESS;
    cellHeap: CARDINAL;
    BEGIN
      cellHeap:=3*q.ndxLen;
      ALLOCATE(cell,cellHeap);

      Seek(h,q.ndxPos+q.ndxOfs + VAL(LONGCARD,ndxLoc)*VAL(LONGCARD,q.ndxLen));Read(h,cell^,cellHeap);

      i:=cell;
      GetNdx(i,q.ndxSize,cell1.lkey,cell1.hkey,cell1.pos,cell1.ofs,cell1.n);IncAddr(i,q.ndxLen);
      GetNdx(i,q.ndxSize,cell2.lkey,cell2.hkey,cell2.pos,cell2.ofs,cell2.n);IncAddr(i,q.ndxLen);
      GetNdx(i,q.ndxSize,cell3.lkey,cell3.hkey,cell3.pos,cell3.ofs,cell3.n);

      cell1.ndxLoc:=ndxLoc;
      cell2.ndxLoc:=ndxLoc+1;
      cell3.ndxLoc:=ndxLoc+2;

      DEALLOCATE(cell,cellHeap)
    END RdCell3;

    PROCEDURE Select;

      PROCEDURE Join(): BOOLEAN;
      VAR
      pagMax: LONGCARD;

        PROCEDURE SetMin;
        BEGIN
          IF cell1.n < cell3.n THEN
             lcell:=cell1;
             hcell:=cell2
          ELSE
             lcell:=cell2;
             hcell:=cell3
          END
        END SetMin;

      BEGIN
        pagMax:=VAL(LONGCARD,q.pagMax)-2;

           IF (cell2.n+cell1.n < pagMax) AND (cell2.n+cell3.n < pagMax) THEN SetMin;                   RETURN TRUE
        ELSIF  cell2.n+cell1.n < pagMax                                 THEN lcell:=cell1;hcell:=cell2;RETURN TRUE
        ELSIF                                 cell2.n+cell3.n < pagMax  THEN lcell:=cell2;hcell:=cell3;RETURN TRUE
         ELSE                                                                                                  RETURN FALSE
          END
      END Join;

    BEGIN
      IF Join() THEN
         RETURN
      END;

      (*combine pages*)
      IF cell1.n > cell3.n THEN
         lcell:=cell1;
         hcell:=cell2
      ELSE
         lcell:=cell2;
         hcell:=cell3
      END
    END Select;

    PROCEDURE Do;
    VAR
    pag,i              : ADDRESS;
    pagHeap,lHeap,hHeap: CARDINAL;

      PROCEDURE JoinPages;
      VAR
      ofs: LONGCARD;
      BEGIN
        ofs:=(q.PAGHEAP - VAL(LONGCARD,pagHeap)) DIV 2;

        Seek(h,lcell.pos+ofs);WrBin(h,pag^,pagHeap);

        WrNdx(h,q.ndxPos,q.ndxOfs,lcell.ndxLoc,q.ndxSize,lcell.lkey,hcell.hkey,lcell.pos,ofs,lcell.n+hcell.n);

        ndxLoc:=hcell.ndxLoc;RemoveNdx
      END JoinPages;

      PROCEDURE CombinePages;
      VAR
      ndx             : ADDRESS;
      ndxHeap         : CARDINAL;
      total,low,high  : CARDINAL;
      lowHeap,highHeap: CARDINAL;
      lowOfs,highOfs,_: LONGCARD;
      _lkey,_hkey     : KEYITEM;
      BEGIN
        ndxHeap:=2*q.ndxLen;
        ALLOCATE(ndx,ndxHeap);

        total   :=VAL(CARDINAL,lcell.n)+VAL(CARDINAL,hcell.n);
        low     :=total DIV 2;
        high    :=total-low;

         lowHeap:= low*q.pagLen;
        highHeap:=high*q.pagLen;

         lowOfs :=(q.PAGHEAP - VAL(LONGCARD, lowHeap)) DIV 2;
        highOfs :=(q.PAGHEAP - VAL(LONGCARD,highHeap)) DIV 2;

        GetPag(AddAddr(pag,(low-1)*q.pagLen),q.ndxSize,_hkey,_);
        GetPag(AddAddr(pag, low   *q.pagLen),q.ndxSize,_lkey,_);

        i:=pag;
        Seek(h,lcell.pos+ lowOfs);WrBin(h,i^, lowHeap);IncAddr(i,lowHeap);
        Seek(h,hcell.pos+highOfs);WrBin(h,i^,highHeap);

        i:=ndx;
        SetNdx(i,q.ndxSize,lcell.lkey,     _hkey,lcell.pos, lowOfs,VAL(LONGCARD,low ));IncAddr(i,q.ndxLen);
        SetNdx(i,q.ndxSize,     _lkey,hcell.hkey,hcell.pos,highOfs,VAL(LONGCARD,high));

        Seek(h,q.ndxPos+q.ndxOfs + VAL(LONGCARD,lcell.ndxLoc)*VAL(LONGCARD,q.ndxLen));WrBin(h,ndx^,ndxHeap);

        DEALLOCATE(ndx,ndxHeap)
      END CombinePages;

    BEGIN
        lHeap:=VAL(CARDINAL,lcell.n)*q.pagLen;
        hHeap:=VAL(CARDINAL,hcell.n)*q.pagLen;
      pagHeap:=lHeap+hHeap;

      ALLOCATE(pag,pagHeap);

      i:=pag;
      Seek(h,lcell.pos+lcell.ofs);Read(h,i^,lHeap);IncAddr(i,lHeap);
      Seek(h,hcell.pos+hcell.ofs);Read(h,i^,hHeap);

      IF VAL(CARDINAL,lcell.n)+VAL(CARDINAL,hcell.n)<q.pagMax-2 THEN
         JoinPages
      ELSE
         CombinePages
      END;

      DEALLOCATE(pag,pagHeap)
    END Do;

  BEGIN
       IF q.n < 3                          THEN RETURN
    ELSIF VAL(CARDINAL,n) > q.pagMax DIV 2 THEN RETURN
    ELSIF ndxLoc=0                         THEN RdCell2(ndxLoc  ,lcell ,hcell        )
    ELSIF ndxLoc=q.n-1                     THEN RdCell2(ndxLoc-1,lcell ,hcell        )
     ELSE                                       RdCell3(ndxLoc-1, cell1, cell2, cell3);Select
      END;

    Do
  END Combine;

BEGIN
  (*main*)
  index:=main;
  IF NOT QCommon(id,h,q,index,FALSE,TRUE) THEN
     RETURN
  END;

  Make(ADR(key),ADR(dat),q.make);

  CASE QSearch(q,key,pos,ofs,n,ndxLoc,TRUE) OF
     foundBottom  : search:=foundBottom;found:=TRUE;FoundBottom
    |foundTop     : search:=foundTop   ;found:=TRUE;FoundTop
    |foundThis    : search:=foundThis  ;            FoundThis
    |notFoundFirst: Unlock(q.h);RETURN
    |notFoundLast : Unlock(q.h);RETURN
    |notFound     : Unlock(q.h);RETURN
  ELSE
                    Error("bt/Delete/1")
  END;

     IF NOT found          THEN       Unlock(q.h);RETURN
  ELSIF (n=1) AND (q.n=1)  THEN Reset;Unlock(q.h);RETURN
  ELSIF  n=1               THEN RemoveNdx
  ELSIF search=foundBottom THEN RemoveBottom;Combine
  ELSIF search=foundTop    THEN RemoveTop   ;Combine
  ELSIF search=foundThis   THEN RemoveThis  ;Combine
    END;

  (*seq*)
  index:=seq;
  IF NOT QCommon(id,h,q,index,FALSE,FALSE) THEN
     Error("bt/Delete/2")
  END;

  Make(ADR(key),ADR(dat),q.make);

  CASE QSearch(q,key,pos,ofs,n,ndxLoc,TRUE) OF
     foundBottom  : search:=foundBottom;found:=TRUE
    |foundTop     : search:=foundTop   ;found:=TRUE
    |foundThis    : search:=foundThis  ;            FoundThis
    |notFoundFirst: Error("bt/Delete/3")
    |notFoundLast : Error("bt/Delete/4")
    |notFound     : Error("bt/Delete/5")
  ELSE
                    Error("bt/Delete/6")
  END;

     IF NOT found          THEN Error("bt/Delete/7")
  ELSIF (n=1) AND (q.n=1)  THEN Error("bt/Delete/8")
  ELSIF  n=1               THEN RemoveNdx
  ELSIF search=foundBottom THEN RemoveBottom;Combine
  ELSIF search=foundTop    THEN RemoveTop   ;Combine
  ELSIF search=foundThis   THEN RemoveThis  ;Combine
    END;

  (*aux*)
  IF NOT q.aux THEN
     Unlock(q.h);RETURN
  END;

  index:=aux;
  IF NOT QCommon(id,h,q,index,FALSE,FALSE) THEN
     Error("bt/Delete/9")
  END;

  Make(ADR(key),ADR(dat),q.make);

  CASE QSearch(q,key,pos,ofs,n,ndxLoc,TRUE) OF
     foundBottom  : search:=foundBottom;found:=TRUE
    |foundTop     : search:=foundTop   ;found:=TRUE
    |foundThis    : search:=foundThis  ;            FoundThis
    |notFoundFirst: Error("bt/Delete/10" )
    |notFoundLast : Error("bt/Delete/11")
    |notFound     : Error("bt/Delete/12")
  ELSE
                    Error("bt/Delete/13")
  END;

     IF NOT found          THEN Error("bt/Delete/14")
  ELSIF (n=1) AND (q.n=1)  THEN Error("bt/Delete/15")
  ELSIF  n=1               THEN RemoveNdx
  ELSIF search=foundBottom THEN RemoveBottom;Combine
  ELSIF search=foundTop    THEN RemoveTop   ;Combine
  ELSIF search=foundThis   THEN RemoveThis  ;Combine
    END;

  Unlock(q.h)
END Delete;

PROCEDURE Info(id: LONGCARD);
VAR
head : HEAD;
h,dev: FIO.File;

  PROCEDURE Key(key    : KEY;
                pagHeap: LONGCARD;
                title  : ARRAY OF CHAR);
  VAR
  len,max,pagLen,pagMax: LONGCARD;
  BEGIN
    len   :=VAL(LONGCARD,NdxLen(VAL(CARDINAL,key.size)));
    max   :=NDXHEAP DIV len;
    pagLen:=VAL(LONGCARD,PagLen(VAL(CARDINAL,key.size)));
    pagMax:=pagHeap DIV pagLen;

    FIO.WrStr(dev,title);FIO.WrLn(dev);

    FIO.WrStr(dev,"size...:");FIO.WrLngCard(dev,key.size     ,10);FIO.WrLn(dev);
    FIO.WrStr(dev,"recMax.:");FIO.WrLngCard(dev,   (   max-2)
                                                  *(pagMax-2),10);FIO.WrLn(dev)
  END Key;

BEGIN
  h:=File(id)^.h;
  Lock(h);Seek(h,0);Read(h,head,SIZE(head));

  (*write info*)
  dev:=FIO.StandardOutput;

  FIO.WrStr(dev,"pagHeap:");FIO.WrLngCard(dev,head.pagHeap,10);FIO.WrLn(dev);
  FIO.WrStr(dev,"recSize:");FIO.WrLngCard(dev,head.recSize,10);FIO.WrLn(dev);
  FIO.WrStr(dev,"recs...:");FIO.WrLngCard(dev,head.n      ,10);FIO.WrLn(dev);

  FIO.WrLn(dev);Key(head.seq,     SEQHEAP,"SEQ  KEY");
  FIO.WrLn(dev);Key(head.ndx,head.pagHeap,"MAIN KEY");
  IF head.aux.size#0 THEN
  FIO.WrLn(dev);Key(head.aux,head.pagHeap,"AUX  KEY")
  END;

  Unlock(h)
END Info;

BEGIN
  FIO.ShareMode:=FIO.ShareDenyNone;

  idx:=ADR(idxDat);

  firma[0]:=225;
  firma[1]:=226;
  firma[2]:=250;
  firma[3]:=26;

  seqMake[0].type:="Q";seqMake[1].type:=KEYEND;
  seqMake[0].pos :=0;
  seqMake[0].data:=4;
  seqMake[0].size:=SEQSIZE;

  _c:=BLANK;

  Reset
END bt.
