(*BT DEMO
  Fredi Cobo, Software Consultant
  fredcobo@yahoo.com*)


(*   DOS: TS 3.10 directives, set project to extended or xlarge
 *
 *   (*$S+*)    (*Stack overflow checking               *)
 *   (*$O+*)    (*Whole numbers overflow checking       *)
 *   (*$I+*)    (*Index checking                        *)
 *   (*$Z+*)    (*Dereferencing of NIL pointers checking*)
 *   (*$S 8000*)(*Stack size: 32k                       *)
 *)


(*   WINDOWS: XDS 2.45 directives                       *)

<*+M2EXTENSIONS*>
<*+M2ADDTYPES*>
<*ALIGNMENT="1"*>

MODULE Demo;

FROM Str IMPORT Copy;
IMPORT IO,bt;

TYPE
REC= RECORD
       seq     : LONGCARD;
       customer: ARRAY[0..2] OF CHAR;
       month   : LONGCARD;
       day     : LONGCARD;
       value   : LONGREAL
     END;

(*
Main key: customer + month + day + seq
Aux  key: month + day + seq
*)

PROCEDURE Wait;
VAR
s: ARRAY[0..24] OF CHAR;
BEGIN
  IO.WrStr("press any key to continue...");IO.RdStr(s)
END Wait;

PROCEDURE Create;
VAR
head: bt.HEAD;
BEGIN
  head.pagHeap         :=8192;
  head.recSize         :=SIZE(REC);
  head.fixSize         :=TRUE;

  (*main key*)
  head.ndx.make[0].type:="S";             (*type of the field 'customer'      *)
  head.ndx.make[0].pos :=4;               (*position of the field in th record*)
  head.ndx.make[0].data:=2;               (*length of the data field          *)
  head.ndx.make[0].size:=2;               (*bytes used to make the key        *)

  head.ndx.make[1].type:="L";             (*field 'month'                     *)
  head.ndx.make[1].pos :=7;
  head.ndx.make[1].data:=4;
  head.ndx.make[1].size:=4;

  head.ndx.make[2].type:="L";             (*field 'day'                       *)
  head.ndx.make[2].pos :=11;
  head.ndx.make[2].data:=4;
  head.ndx.make[2].size:=4;

  head.ndx.make[3].type:="Q";             (*field 'seq'                       *)
  head.ndx.make[3].pos :=0;
  head.ndx.make[3].data:=4;
  head.ndx.make[3].size:=3;

  head.ndx.make[4].type:=bt.KEYEND;       (*end of key                        *)

  (*aux key*)
  head.aux.make[0].type:="L";             (*field 'month'                     *)
  head.aux.make[0].pos :=7;
  head.aux.make[0].data:=4;
  head.aux.make[0].size:=4;

  head.aux.make[1].type:="L";             (*field 'day'                       *)
  head.aux.make[1].pos :=11;
  head.aux.make[1].data:=4;
  head.aux.make[1].size:=4;

  head.aux.make[2].type:="Q";             (*field 'seq'                       *)
  head.aux.make[2].pos :=0;
  head.aux.make[2].data:=4;
  head.aux.make[2].size:=3;

  head.aux.make[3].type:=bt.KEYEND;

  bt.Create("tra.bt",head,0)
END Create;

PROCEDURE WriteData;
VAR
h: LONGCARD;

  PROCEDURE Add(customer : ARRAY OF CHAR;
                month,day: LONGCARD;
                value    : LONGREAL);
  VAR
  r: REC;
  BEGIN
    Copy(r.customer,customer);

    r.month :=month;
    r.day   :=day;
    r.value :=value;

    IF NOT bt.Add(h,r) THEN         (*the 'seq' field is given a value*)
       IO.WrStr("add error");HALT   (*inside Add()                    *)
    END
  END Add;

BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  Add("JK", 3, 4,17.42);
  Add("LS", 2,17, 2.45);
  Add("KH", 1,12, 1.18);
  Add("KH", 5, 5,21.45);
  Add("LS", 2,13, 1.12);
  Add("LS", 4,24, 1.45);
  Add("JK", 1,23, 2.12);
  Add("LS", 2, 6, 4.03);
  Add("LS", 2,17, 2.32);

  bt.Close(h)
END WriteData;

PROCEDURE RetrieveSingleRecord;
VAR
h: LONGCARD;
r: REC;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  r.seq:=4;
  IF bt.Find(h,r,bt.seq) THEN
     IO.WrLn;
     IO.WrStr       (r.customer    );
     IO.WrLngCard   (r.month   ,  4);
     IO.WrLngCard   (r.day     ,  4);
     IO.WrFixLngReal(r.value   ,2,8);
     IO.WrLn
  ELSE
     IO.WrLn;
     IO.WrStr("not found.");
     IO.WrLn
  END;

  bt.Close(h)
END RetrieveSingleRecord;

PROCEDURE RetrieveAllRecords(index: bt.IndexType);
VAR
h: LONGCARD;
q: bt.QUERY;
r: REC;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  bt.Query(h,q,index);

  (*you have the number of records of the query in advance*)
  IO.WrLn;
  IO.WrStr("records: ");IO.WrLngCard(q.recs,2);IO.WrLn;

  WHILE bt.Next(q,r) DO
    IO.WrLngCard   (r.seq     ,  2);
    IO.WrStr       (  "  "        );
    IO.WrStr       (r.customer    );
    IO.WrLngCard   (r.month   ,  4);
    IO.WrLngCard   (r.day     ,  4);
    IO.WrFixLngReal(r.value   ,2,8);
    IO.WrLn
  END;

  bt.Close(h);
END RetrieveAllRecords;

PROCEDURE RetrieveSubkey;
VAR
h: LONGCARD;
q: bt.QUERY;
r: REC;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  r.customer:="KH";                       (*subkey of main key, 2 bytes length*)
  bt.Query2(h,q,r,2,bt.main);

  IO.WrLn;IO.WrStr("records: ");IO.WrLngCard(q.recs,2);IO.WrLn;

  WHILE bt.Next(q,r) DO
    IO.WrLngCard   (r.seq     ,  2);
    IO.WrStr       (  "  "        );
    IO.WrStr       (r.customer    );
    IO.WrLngCard   (r.month   ,  4);
    IO.WrLngCard   (r.day     ,  4);
    IO.WrFixLngReal(r.value   ,2,8);
    IO.WrLn
  END;

  bt.Close(h)
END RetrieveSubkey;

PROCEDURE RetrieveSubkeyRange;
VAR
h      : LONGCARD;
q      : bt.QUERY;
r1,r2,r: REC;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  r1.month:=2;                      (*first  subkey of aux key, 4 bytes length*)
  r2.month:=4;                      (*second subkey of aux key, 4 bytes length*)

  bt.Query3(h,q,r1,r2,4,bt.aux);

  IO.WrLn;IO.WrStr("records: ");IO.WrLngCard(q.recs,2);IO.WrLn;

  WHILE bt.Next(q,r) DO
    IO.WrLngCard   (r.seq     ,  2);
    IO.WrStr       (  "  "        );
    IO.WrStr       (r.customer    );
    IO.WrLngCard   (r.month   ,  4);
    IO.WrLngCard   (r.day     ,  4);
    IO.WrFixLngReal(r.value   ,2,8);
    IO.WrLn
  END;

  bt.Close(h)
END RetrieveSubkeyRange;

PROCEDURE ChangeRecord;
VAR
h: LONGCARD;
r: REC;

  PROCEDURE Echo(s: ARRAY OF CHAR);
  BEGIN
    IO.WrStr       (  s           );
    IO.WrStr       (r.customer    );
    IO.WrLngCard   (r.month   ,  4);
    IO.WrLngCard   (r.day     ,  4);
    IO.WrFixLngReal(r.value   ,2,8);
    IO.WrLn
  END Echo;

BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  r.seq:=5;
  IF NOT bt.Find(h,r,bt.seq) THEN
     bt.Close(h);IO.WrStr("not found");RETURN
  END;

  IO.WrLn;
  Echo("before: ");

  r.value:=16.87;
  IF NOT bt.Change(h,r) THEN
     bt.Close(h);IO.WrStr("change error");HALT
  END;

  Echo("after : ");

  bt.Close(h)
END ChangeRecord;

PROCEDURE DeleteRecord;
VAR
h: LONGCARD;
r: REC;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  r.seq:=5;
  IF NOT bt.Find(h,r,bt.seq) THEN
     bt.Close(h);IO.WrStr("not found");RETURN
  END;

  bt.Delete(h,r);

  bt.Close(h)
END DeleteRecord;

PROCEDURE Information;
VAR
h: LONGCARD;
BEGIN
  h:=bt.Open("tra.bt",SIZE(REC));

  IO.WrLn;
  bt.Info(h);

  bt.Close(h)
END Information;

BEGIN
  IO.WrStr("Creating structure for 'tra.bt'"                    );Create;                          IO.WrStr(".");IO.WrLn;
  IO.WrStr("Writing data"                                       );WriteData;                       IO.WrStr(".");IO.WrLn;IO.WrLn;

  IO.WrStr("Retrieve single record, seq=4"                      );RetrieveSingleRecord;            IO.WrLn;

  IO.WrStr("Retrieve all records, seq key"                      );RetrieveAllRecords(bt.seq );Wait;IO.WrLn;
  IO.WrStr("Retrieve all records, main key"                     );RetrieveAllRecords(bt.main);Wait;IO.WrLn;
  IO.WrStr("Retrieve all records, aux key"                      );RetrieveAllRecords(bt.aux );Wait;IO.WrLn;

  IO.WrStr("Retrieve subkey of main key, customer 'KH'"         );RetrieveSubkey;             Wait;IO.WrLn;

  IO.WrStr("Retrieve subkey range of aux key, february to april");RetrieveSubkeyRange;        Wait;IO.WrLn;

  IO.WrStr("Change record, seq=5"                               );ChangeRecord;               Wait;IO.WrLn;

  IO.WrStr("Delete record, seq=5"                               );DeleteRecord;                   ;IO.WrLn;
  IO.WrStr("Retrieve all records, seq key"                      );RetrieveAllRecords(bt.seq );Wait;IO.WrLn;

  IO.WrStr("Information for 'tra.bt'"                           );Information
END Demo.
